"""
This module provides tools for creating kmz products for a SICD type element.

.. Note::
    Creation of ground overlays (i.e. image overlay) requires the optional
    Pillow dependency for image manipulation.

Examples
--------
Create a kmz overview for the contents of a sicd type reader.

.. code-block:: python

    import os
    from sarpy.io.complex.converter import open_complex
    from sarpy.visualization.kmz_product_creation import create_kmz_view

    test_root = '<root directory>'
    reader = open_complex(os.path.join(test_root, '<file name>>'))
    create_kmz_view(reader, test_root,
                    file_stem='View-<something descriptive>',
                    pixel_limit=2048,
                    inc_collection_wedge=True)
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"

# TODO: tidy up significantly


import logging
from typing import Union
import json
import os

import numpy

from sarpy.processing.rational_polynomial import SarpyRatPolyError
from sarpy.processing.ortho_rectify.base import FullResolutionFetcher, OrthorectificationIterator
from sarpy.processing.ortho_rectify.ortho_methods import OrthorectificationHelper, NearestNeighborMethod
from sarpy.processing.ortho_rectify.projection_helper import PGProjection, PGRatPolyProjection
from sarpy.io.kml import Document
from sarpy.io.complex.base import SICDTypeReader
from sarpy.io.complex.sicd_elements.SICD import SICDType
from sarpy.io.complex.sicd_elements.blocks import Poly2DType
from sarpy.io.complex.utils import sicd_reader_iterator
from sarpy.geometry.geocoords import ecf_to_geodetic
from sarpy.visualization import kmz_utils
from sarpy.visualization.remap import RemapFunction, NRL

try:
    # noinspection PyPackageRequirements
    import PIL
    import PIL.Image
except ImportError:
    PIL = None

logger = logging.getLogger(__name__)


def _create_sicd_styles(kmz_document):
    """
    Creates the appropriate styles for SICD usage.

    Parameters
    ----------
    kmz_document : Document

    Returns
    -------
    None
    """
    kmz_utils.add_antenna_styles(kmz_document)

    # bounding box style - maybe polygon, maybe corner points, clamped to ground
    label = {'color': 'ffc0c0c0', 'scale': '1.0'}
    icon = {'scale': '1.5', 'icon_ref': 'http://maps.google.com/mapfiles/kml/pushpin/blue-pushpin.png'}
    line = {'color': 'ccff5050', 'width': '2.0'}
    poly = {'color': '30ff5050'}
    kmz_document.add_style('bounding_high', label_style=label, icon_style=icon, line_style=line, poly_style=poly)
    label['scale'] = '0.75'
    icon['scale'] = '1.0'
    line['width'] = '1.0'
    kmz_document.add_style('bounding_low', label_style=label, icon_style=icon, line_style=line, poly_style=poly)
    kmz_document.add_style_map('bounding', 'bounding_high', 'bounding_low')

    # valid data style - basic polygon, probably clamped to ground
    line = {'color': 'cc5050ff', 'width': '2.0'}
    poly = {'color': '305050ff'}
    kmz_document.add_style('valid_high', line_style=line, poly_style=poly)
    line['width'] = '1.0'
    kmz_document.add_style('valid_low', line_style=line, poly_style=poly)
    kmz_document.add_style_map('valid', 'valid_high', 'valid_low')

    # scp - intended for basic point clamped to ground
    label = {'color': 'ff50c0c0', 'scale': '1.0'}
    icon = {'color': 'ff5050c0', 'scale': '1.5',
            'icon_ref': 'http://maps.google.com/mapfiles/kml/shapes/shaded_dot.png'}
    kmz_document.add_style('scp_high', label_style=label, icon_style=icon)
    label['scale'] = '0.75'
    icon['scale'] = '1.0'
    kmz_document.add_style('scp_low', label_style=label, icon_style=icon)
    kmz_document.add_style_map('scp', 'scp_high', 'scp_low')

    # arp position style - intended for gx track
    line = {'color': 'ff50ff50', 'width': '1.5'}
    label = {'color': 'ffc0c0c0', 'scale': '1.5'}
    icon = {'scale': '2.0', 'icon_ref': 'http://maps.google.com/mapfiles/kml/shapes/track.png'}
    poly = {'color': 'a050ff50'}
    kmz_document.add_style('arp_high', line_style=line, label_style=label, icon_style=icon, poly_style=poly)
    line['width'] = '1.0'
    label['scale'] = '1.0'
    icon['scale'] = '1.0'
    poly = {'color': '7050ff50'}
    kmz_document.add_style('arp_low', line_style=line, label_style=label, icon_style=icon, poly_style=poly)
    kmz_document.add_style_map('arp', 'arp_high', 'arp_low')

    # collection wedge style - intended as polygon
    line = {'color': 'ffa0a050', 'width': '1.5'}
    poly = {'color': 'a0a0a050'}
    kmz_document.add_style('collection_high', line_style=line, poly_style=poly)
    line['width'] = '1.0'
    poly = {'color': '70a0a050'}
    kmz_document.add_style('collection_low', line_style=line, poly_style=poly)
    kmz_document.add_style_map('collection', 'collection_high', 'collection_low')


def _get_sicd_name(sicd):
    """
    Gets the kml-styled name for the provided SICD.

    Parameters
    ----------
    sicd : SICDType

    Returns
    -------
    str
    """

    return sicd.CollectionInfo.CoreName


def _get_sicd_description(sicd):
    """
    Gets the kml-styled description for the provided SICD.

    Parameters
    ----------
    sicd : SICDType

    Returns
    -------
    str
    """

    o_sicd = sicd.copy()
    # junk the WgtFunct, it's huge and probably not interesting
    try:
        o_sicd.Grid.Row.WgtFunct = None
        o_sicd.Grid.Col.WgtFunct = None
    except AttributeError:
        pass

    return json.dumps(o_sicd.to_dict(), indent=1)


def _get_orthoiterator_description(ortho_iterator):
    """
    Get a description for the ortho_iterator details.

    Parameters
    ----------
    ortho_iterator : OrthorectificationIterator

    Returns
    -------
    str
    """

    return 'ortho-rectified image for {2:s}<br>' \
           'row resolution - {0:0.2f} meters<br>' \
           'column resolution - {1:0.2f} meters<br>' \
           'remap function - {3:s}'.format(
        ortho_iterator.ortho_helper.proj_helper.row_spacing,
        ortho_iterator.ortho_helper.proj_helper.col_spacing,
        _get_sicd_name(ortho_iterator.sicd),
        ortho_iterator.remap_function.name)


def _get_sicd_time_args(sicd, subdivisions=24):
    # type: (SICDType, Union[int, None]) -> (dict, Union[None, numpy.ndarray])
    """
    Fetch the SICD time arguments and array.

    Parameters
    ----------
    sicd : SICDType
    subdivisions : int|None

    Returns
    -------
    (dict, None|numpy.ndarray)
    """

    if sicd.Timeline is None or sicd.Timeline.CollectStart is None:
        return {}, None

    col_start = sicd.Timeline.CollectStart.astype('datetime64[us]')
    if sicd.ImageFormation.TStartProc is None:
        return {'when': str(col_start) + 'Z', }, None

    t_start = sicd.ImageFormation.TStartProc
    t_end = sicd.ImageFormation.TEndProc
    if not isinstance(subdivisions, int) or subdivisions < 2:
        time_array = None
    else:
        time_array = numpy.linspace(t_start, t_end, subdivisions)

    beg_time = col_start + int(t_start * 1e6)
    end_time = col_start + int(t_end * 1e6)
    return {'beginTime': str(beg_time) + 'Z', 'endTime': str(end_time) + 'Z'}, time_array


def _write_image_corners(kmz_document, sicd, time_args, folder):
    """
    Write the image corner.

    Parameters
    ----------
    kmz_document : Document
    sicd : SICDType
    time_args : dict
    folder : minidom.Element

    Returns
    -------
    None
    """

    if sicd.GeoData is None or sicd.GeoData.ImageCorners is None:
        return

    frm = '{1:0.8f},{0:0.8f},0'
    corners = sicd.GeoData.ImageCorners.get_array(dtype='float64')

    if numpy.any(~numpy.isfinite(corners)):
        logger.error('There are nonsense entries (nan or +/- infinity) in the corner locations array.')

    names = ['FRFC', 'FRLC', 'LRLC', 'LRFC']
    for nam, corner in zip(names, corners):
        if numpy.any(~numpy.isfinite(corner)):
            continue
        coords = frm.format(*corner)
        placemark = kmz_document.add_container(par=folder, description='{} for {}'.format(nam, _get_sicd_name(sicd)),
                                               styleUrl='#bounding')
        kmz_document.add_point(coords, par=placemark, altitudeMode='clampToGround', **time_args)


def _write_valid_area(kmz_document, sicd, time_args, folder):
    """
    Write the valid area polygon.

    Parameters
    ----------
    kmz_document : Document
    sicd : SICDType
    time_args : dict
    folder : minidom.Element

    Returns
    -------
    None
    """

    if sicd.GeoData is None or sicd.GeoData.ValidData is None:
        return

    frm = '{1:0.8f},{0:0.8f},0'
    valid_array = sicd.GeoData.ValidData.get_array(dtype='float64')
    if numpy.any(~numpy.isfinite(valid_array)):
        logger.error('There are nonsense entries (nan or +/- infinity) in the valid array location.')

    coords = ' '.join(frm.format(*el) for el in valid_array)
    coords += ' ' + frm.format(*valid_array[0, :])
    placemark = kmz_document.add_container(par=folder, description='valid data for {}'.format(_get_sicd_name(sicd)),
                                           styleUrl='#valid')
    kmz_document.add_polygon(coords, par=placemark, altitudeMode='clampToGround', **time_args)


def _write_scp(kmz_document, sicd, time_args, folder):
    """
    Write the scp location.

    Parameters
    ----------
    kmz_document : Document
    sicd : SICDType
    time_args : dict
    folder : minidom.Element

    Returns
    -------
    None
    """
    if sicd.GeoData is None or sicd.GeoData.SCP is None:
        return

    scp_llh = sicd.GeoData.SCP.LLH.get_array()
    if numpy.any(~numpy.isfinite(scp_llh)):
        logger.error('There are nonsense entries (nan or +/- infinity) in the scp location.')

    frm = '{1:0.8f},{0:0.8f},0'
    coords = frm.format(*scp_llh)
    placemark = kmz_document.add_container(par=folder, description='SCP for {}'.format(_get_sicd_name(sicd)),
                                           styleUrl='#scp')
    kmz_document.add_point(coords, par=placemark, altitudeMode='clampToGround', **time_args)


def _write_arp_location(kmz_document, sicd, time_args, time_array, folder):
    """

    Parameters
    ----------
    kmz_document : Document
    sicd : SICDType
    time_args : dict
    time_array : None|numpy.ndarray
    folder : minidom.Element

    Returns
    -------
    None|Numpy.ndarray
    """

    if time_array is None:
        return None

    if sicd.Position is not None and sicd.Position.ARPPoly is not None:
        arp_pos = sicd.Position.ARPPoly(time_array)
    elif sicd.SCPCOA.ARPPos is not None and sicd.SCPCOA.ARPVel is not None:
        arp_pos = sicd.SCPCOA.ARPPos.get_array() + numpy.outer(time_array, sicd.SCPCOA.ARPVel.get_array())
    else:
        return None

    arp_llh = ecf_to_geodetic(arp_pos)
    if numpy.any(~numpy.isfinite(arp_llh)):
        logger.error('There are nonsense entries (nan or +/- infinity) in the aperture location.')
    coords = ['{1:0.8f},{0:0.8f},{2:0.2f}'.format(*el) for el in arp_llh]
    whens = [str(sicd.Timeline.CollectStart.astype('datetime64[us]') + int(el * 1e6)) + 'Z' for el in time_array]
    placemark = kmz_document.add_container(par=folder, description='aperture position for {}'.format(_get_sicd_name(
        sicd)), styleUrl='#arp', **time_args)
    kmz_document.add_gx_track(coords, whens, par=placemark, extrude=True, tesselate=True, altitudeMode='absolute')
    return arp_llh


def _write_collection_wedge(kmz_document, sicd, time_args, arp_llh, time_array, folder):
    """
    Writes the collection wedge.

    Parameters
    ----------
    kmz_document : Document
    sicd : SICDType
    time_args : dict
    arp_llh : None|numpy.ndarray
    time_array : None|numpy.ndarray
    folder : minidom.Element

    Returns
    -------
    None
    """

    if time_array is None or arp_llh is None:
        return

    if sicd.Position is not None and sicd.Position.GRPPoly is not None:
        grp = sicd.Position.GRPPoly(time_array)
    elif sicd.GeoData is not None and sicd.GeoData.SCP is not None:
        grp = numpy.reshape(sicd.GeoData.SCP.ECF.get_array(), (1, 3))
    else:
        return
    grp_llh = ecf_to_geodetic(grp)

    if numpy.any(~numpy.isfinite(grp_llh)):
        logger.error('There are nonsense entries (nan or +/- infinity) in the scp/ground range locations.')

    coord_array = numpy.concatenate([arp_llh, grp_llh[::-1]], axis=0)
    coord_array[:, [0, 1]] = coord_array[:, [1, 0]]  # lat,lon -> lon,lat
    coords = " ".join(",".join(str(x) for x in coord) for coord in coord_array)
    placemark = kmz_document.add_container(par=folder, description='collection wedge for {}'.format(_get_sicd_name(
        sicd)), styleUrl='#collection', **time_args)
    kmz_document.add_polygon(coords, par=placemark, extrude=False, tesselate=False, altitudeMode='absolute')


def _write_antenna(kmz_document, sicd, time_args, time_array, folder):
    """
    Write the Antenna information

    Parameters
    ----------
    kmz_document : Document
    sicd : SICDType
    time_args : dict
    time_array : None|numpy.ndarray
    folder : minidom.Element

    Returns
    -------
    None
    """
    if time_array is None:
        return

    if sicd.Antenna is None:
        return

    boresight_folder = kmz_document.add_container(
        par=folder,
        the_type="Folder",
        name="Boresights",
    )

    footprint_folder = kmz_document.add_container(
        par=folder,
        the_type="Folder",
        name="Beam Footprints",
    )

    contour_levels = {
        'Tx': -3,
        'Rcv': -3,
        'TwoWay': -6,
    }

    for which in ("Tx", "Rcv", "TwoWay"):
        aiming = antenna_aiming(sicd, which, time_array)
        if not aiming:
            continue

        for boresight_type in ("mechanical", "electrical"):
            visibility = which == "Rcv"  # only display Rcv by default
            name = f"{which} {boresight_type} boresight"

            on_earth_ecf = numpy.asarray(
                [
                    kmz_utils.ray_intersect_earth(apc_pos, along)
                    for apc_pos, along in zip(
                    aiming["raw"]["positions"],
                    aiming[boresight_type],
                )
                ]
            )

            placemark = kmz_document.add_container(
                par=boresight_folder,
                name=name,
                description=f"{name} <br><br>Highlighted edge indicates start time",
                styleUrl=f"#{boresight_type}_boresight",
                visibility=visibility,
                **time_args,
            )
            boresight_coords = kmz_utils.ecef_to_kml_coord(on_earth_ecf)
            apc_coords = kmz_utils.ecef_to_kml_coord(aiming["raw"]["positions"])

            kmz_utils.add_los_polygon(kmz_document,
                                      placemark,
                                      apc_coords,
                                      boresight_coords,
                                      )
            footprint_labels = {
                "start": 0,
                "middle": len(time_array) // 2,
                "end": len(time_array) - 1,
            }

        array_gain_poly = aiming["raw"]["pattern"].Array.GainPoly
        if aiming['raw']['pattern'].Elem is None:
            elem_gain_poly = Poly2DType(Coefs=[[0]])
        else:
            elem_gain_poly = aiming["raw"]["pattern"].Elem.GainPoly

        contour_level = contour_levels[which]
        for when, this_footprint in kmz_utils.make_beam_footprints(
                aiming, footprint_labels, array_gain_poly, elem_gain_poly, contour_level=contour_level,
        ).items():
            name = f"{which} beam {contour_level}dB footprint @ {when}"
            placemark = kmz_document.add_container(
                par=footprint_folder,
                name=name,
                description=name,
                styleUrl=f"#{which.lower()}_beam_footprint",
                visibility=True,
                **time_args,
            )
            coords = kmz_utils.ecef_to_kml_coord(this_footprint["contour"])
            kmz_document.add_polygon(
                " ".join(coords),
                par=placemark,
            )


def _write_sicd_overlay(ortho_iterator, kmz_document, folder):
    """
    Write the orthorectified SICD ground overlay.

    Parameters
    ----------
    ortho_iterator : OrthorectificationIterator
    kmz_document : Document
    folder : minidom.Element

    Returns
    -------
    None
    """

    def reorder_corners(llh_in):
        return llh_in[::-1, :]

    if PIL is None:
        logger.error(
            'This functionality for writing kmz ground overlays requires the optional Pillow dependency.')
        return

    time_args, _ = _get_sicd_time_args(ortho_iterator.sicd, subdivisions=None)

    # create the output workspace
    if ortho_iterator.remap_function.bit_depth != 8:
        raise ValueError('The bit depth for the remap function must be 8, for now.')
    image_data = numpy.zeros(ortho_iterator.ortho_data_size, dtype=ortho_iterator.remap_function.output_dtype)
    # populate by iterating
    for data, start_indices in ortho_iterator:
        image_data[start_indices[0]:start_indices[0] + data.shape[0],
        start_indices[1]:start_indices[1] + data.shape[1]] = data
    # create regionated overlay
    # convert image array to PIL image.
    img = PIL.Image.fromarray(image_data)  # this is to counteract the PIL treatment
    lat_lon_quad = reorder_corners(ortho_iterator.get_llh_image_corners())
    kmz_document.add_regionated_ground_overlay(
        img, folder, lat_lon_quad=lat_lon_quad[:, :2], img_format='JPEG',
        name='image overlay for {}'.format(_get_sicd_name(ortho_iterator.sicd)),
        description=_get_orthoiterator_description(ortho_iterator))


def prepare_kmz_file(file_name, **args):
    """
    Prepare a kmz document and archive for exporting.

    Parameters
    ----------
    file_name : str
    args
        Passed through to the Document constructor.

    Returns
    -------
    Document
    """

    document = Document(file_name=file_name, **args)
    _create_sicd_styles(document)
    return document


def add_sicd_geometry_elements(sicd, kmz_document, folder,
                               inc_image_corners=True, inc_valid_data=False,
                               inc_scp=False, inc_collection_wedge=True, inc_antenna=True):
    """
    Write the geometry elements of a SICD.

    Parameters
    ----------
    sicd : SICDType
    kmz_document : Document
    folder : minidom.Element
    inc_image_corners : bool
        Include the image corners, if possible?
    inc_valid_data : bool
        Include the valid image area, if possible?
    inc_scp : bool
        Include the scp?
    inc_collection_wedge : bool
        Include the aperture location and collection wedge?
    inc_antenna: bool
        Include Antenna footprints and boresights?

    Returns
    -------
    None
    """

    # let's define the time data for the SICD
    time_args, time_array = _get_sicd_time_args(sicd)

    # add the image corners/bounding box
    if inc_image_corners:
        _write_image_corners(kmz_document, sicd, time_args, folder)
    # add the valid data
    if inc_valid_data:
        _write_valid_area(kmz_document, sicd, time_args, folder)
    # write scp data
    if inc_scp:
        _write_scp(kmz_document, sicd, time_args, folder)
    # write arp position and collection wedge
    if inc_collection_wedge:
        arp_llh = _write_arp_location(kmz_document, sicd, time_args, time_array, folder)
        _write_collection_wedge(kmz_document, sicd, time_args, arp_llh, time_array, folder)

    if inc_antenna:
        _write_antenna(kmz_document, sicd, time_args, time_array, folder)


def add_sicd_from_ortho_helper(kmz_document, ortho_helper,
                               inc_image_corners=False, inc_valid_data=False,
                               inc_scp=False, inc_collection_wedge=False,
                               inc_antenna=True,
                               block_size=10, remap_function=None):
    """
    Adds for a SICD to the provided open kmz from an ortho-rectification helper.

    Parameters
    ----------
    kmz_document : Document
    ortho_helper : OrthorectificationHelper
    inc_image_corners : bool
        Include the image corners, if possible?
    inc_valid_data : bool
        Include the valid image area, if possible?
    inc_scp : bool
        Include the scp?
    inc_collection_wedge : bool
        Include the aperture location and collection wedge?
    inc_antenna: bool
        Include Antenna footprints and boresights?
    block_size : None|int|float
        The block size for the iterator
    remap_function : None|RemapFunction
        The remap function to apply, or a suitable default will be chosen.
    """

    if not isinstance(ortho_helper, OrthorectificationHelper):
        raise TypeError(
            'ortho_helper must be an OrthorectificationHelper instance, got '
            'type {}'.format(type(ortho_helper)))
    if not isinstance(kmz_document, Document):
        raise TypeError(
            'kmz_document must be an sarpy.io.kml.Document instance, got '
            'type {}'.format(type(kmz_document)))

    # create a folder for these sicd details
    sicd = ortho_helper.sicd
    folder = kmz_document.add_container(
        the_type='Folder', name=_get_sicd_name(sicd), description=_get_sicd_description(sicd))

    # write the sicd details aside from the overlay
    add_sicd_geometry_elements(sicd, kmz_document, folder,
                               inc_image_corners=inc_image_corners, inc_valid_data=inc_valid_data,
                               inc_scp=inc_scp, inc_collection_wedge=inc_collection_wedge, inc_antenna=inc_antenna)

    # create the ortho-rectification iterator
    if remap_function is None:
        remap_function = NRL()
    calculator = FullResolutionFetcher(
        ortho_helper.reader, index=ortho_helper.index, dimension=1, block_size=block_size)
    ortho_iterator = OrthorectificationIterator(
        ortho_helper, calculator=calculator, remap_function=remap_function,
        recalc_remap_globals=True)

    # write the image overlay
    _write_sicd_overlay(ortho_iterator, kmz_document, folder)


def add_sicd_to_kmz(kmz_document, reader, index=0, pixel_limit=2048,
                    inc_image_corners=False, inc_valid_data=False,
                    inc_scp=False, inc_collection_wedge=False,
                    inc_antenna=True,
                    block_size=10, remap_function=None):
    """
    Adds elements for this SICD to the provided open kmz.

    Parameters
    ----------
    kmz_document : Document
        The kmz document, which must be open and have an associated archive.
    reader : SICDTypeReader
        The reader instance, must be of sicd type:
    index : int
        The index to use.
    pixel_limit : None|int
        The limit in pixel size to use for the constructed ground overlay.
    inc_image_corners : bool
        Include the image corners, if possible?
    inc_valid_data : bool
        Include the valid image area, if possible?
    inc_scp : bool
        Include the scp?
    inc_collection_wedge : bool
        Include the aperture location and collection wedge?
    inc_antenna: bool
        Include Antenna footprints and boresights?
    block_size : None|int|float
        The block size for the iterator
    remap_function : None|RemapFunction
        The remap function to apply, or a suitable default will be chosen.

    Returns
    -------
    None
    """

    if not isinstance(reader, SICDTypeReader):
        raise TypeError('reader must be a instance of SICDTypeReader. Got type {}'.format(type(reader)))

    if pixel_limit is not None:
        pixel_limit = int(pixel_limit)
        if pixel_limit < 512:
            pixel_limit = 512

    # create our projection helper
    index = int(index)
    sicd = reader.get_sicds_as_tuple()[index]
    try:
        proj_helper = PGRatPolyProjection(sicd)
    except SarpyRatPolyError:
        proj_helper = PGProjection(sicd)
    # create our orthorectification helper
    ortho_helper = NearestNeighborMethod(reader, index=index, proj_helper=proj_helper)
    if pixel_limit is not None:
        # let's see what the ortho-rectified size will be
        ortho_size = ortho_helper.get_full_ortho_bounds()
        row_count = ortho_size[1] - ortho_size[0]
        col_count = ortho_size[3] - ortho_size[2]
        # reset the row/column spacing, if necessary
        if row_count > pixel_limit:
            proj_helper.row_spacing *= row_count / float(pixel_limit)
        if col_count > pixel_limit:
            proj_helper.col_spacing *= col_count / float(pixel_limit)
        if isinstance(proj_helper, PGRatPolyProjection):
            proj_helper.perform_rational_poly_fitting()
    # add the sicd details
    add_sicd_from_ortho_helper(
        kmz_document, ortho_helper,
        inc_image_corners=inc_image_corners, inc_valid_data=inc_valid_data, inc_scp=inc_scp,
        inc_collection_wedge=inc_collection_wedge,
        inc_antenna=inc_antenna,
        block_size=block_size, remap_function=remap_function)


def create_kmz_view(
        reader, output_directory, file_stem='view', pixel_limit=2048,
        inc_image_corners=False, inc_valid_data=False,
        inc_scp=True, inc_collection_wedge=False, inc_antenna=True, block_size=10, remap_function=None):
    """
    Create a kmz view for the reader contents. **This will create one file per
    band/polarization present in the reader.**

    Parameters
    ----------
    reader : SICDTypeReader
    output_directory : str
    file_stem : str
    pixel_limit : None|int
    inc_image_corners : bool
        Include the image corners, if possible?
    inc_valid_data : bool
        Include the valid image area, if possible?
    inc_scp : bool
        Include the scp?
    inc_collection_wedge : bool
        Include the aperture location and collection wedge?
    inc_antenna: bool
        Include Antenna footprints and boresights?
    block_size : None|int|float
        The block size for the iterator
    remap_function : None|RemapFunction
        The remap function to apply, or a suitable default will be chosen.

    Returns
    -------
    None

    Examples
    --------
    .. code-block:: python

        import logging
        logger = logging.getLogger('sarpy')
        logger.setLevel('INFO')

        import os
        from sarpy.io.complex.converter import open_complex
        from sarpy.visualization.kmz_product_creation import create_kmz_view

        test_root = '<root directory>'
        reader = open_complex(os.path.join(test_root, '<file name>>'))
        create_kmz_view(reader, test_root,
                        file_stem='View-<something descriptive>',
                        pixel_limit=2048,
                        inc_collection_wedge=True)

    """

    def get_pol_abbreviation(pol_in):
        spol = pol_in.split(':')
        if len(spol) == 2:
            return spol[0][0] + spol[1][0]
        return pol_in

    def do_iteration():
        kmz_file = os.path.join(output_directory, '{}_{}_{}.kmz'.format(file_stem,
                                                                        the_band,
                                                                        get_pol_abbreviation(the_pol)))
        logger.info('Writing kmz file for polarization {} and band {}'.format(the_pol, the_band))
        with prepare_kmz_file(kmz_file, name=reader.file_name) as kmz_doc:
            for the_partition, the_index, the_sicd in sicd_reader_iterator(
                    reader, partitions=partitions, polarization=the_pol, band=the_band):
                add_sicd_to_kmz(
                    kmz_doc, reader,
                    index=the_index, pixel_limit=pixel_limit,
                    inc_image_corners=inc_image_corners, inc_valid_data=inc_valid_data,
                    inc_scp=inc_scp, inc_collection_wedge=inc_collection_wedge,
                    inc_antenna=inc_antenna,
                    block_size=block_size, remap_function=remap_function)

    bands = set(reader.get_sicd_bands())
    pols = set(reader.get_sicd_polarizations())
    partitions = reader.get_sicd_partitions()

    for the_band in bands:
        for the_pol in pols:
            do_iteration()


def antenna_aiming(sicd, which, time_array):
    """
    Compile antenna aiming metadata of a SICD

    Parameters
    ----------
    sicd : SICDType
    which: str
        Which antenna pattern to use: 'Tx', 'Rcv', or 'TwoWay'
    time_array : None|numpy.ndarray
        Points in time to evaluate the antenna related polynomials

    """
    if time_array is None:
        return {}

    if sicd.Antenna is None:
        return {}

    ant_dir_node = getattr(sicd.Antenna, which)
    if ant_dir_node is None:
        return {}

    try:
        if which == 'Tx':
            ant_dir_node = sicd.Antenna.Tx
            apc_poly = sicd.Position.TxAPCPoly
        elif which == 'Rcv':
            ant_dir_node = sicd.Antenna.Rcv
            chan_index = sicd.ImageFormation.RcvChanProc.ChanIndices[0]
            apc_index = sicd.RadarCollection.RcvChannels[chan_index - 1].RcvAPCIndex
            apc_poly = sicd.Position.RcvAPC[apc_index - 1]
        elif which == 'TwoWay':
            ant_dir_node = sicd.Antenna.TwoWay
            apc_poly = sicd.Position.ARPPoly
        else:
            raise ValueError(f"Unknown antenna '{which}'")

        positions = apc_poly(time_array)
        uacx = ant_dir_node.XAxisPoly(time_array)
        uacy = ant_dir_node.YAxisPoly(time_array)
        uacz = numpy.cross(uacx, uacy)

        pointing = {
            "antpat_id": which,
            "raw": {
                "positions": positions,
                "times": time_array,
                "uacx": uacx,
                "uacy": uacy,
                "uacz": uacz,
                "pattern": ant_dir_node,
            },
            "mechanical": uacz,
        }

        if ant_dir_node.EB is None:
            eb_dcx = numpy.zeros(len(time_array))
            eb_dcy = numpy.zeros(len(time_array))
        else:
            eb_dcx = ant_dir_node.EB.DCXPoly(time_array)
            eb_dcy = ant_dir_node.EB.DCYPoly(time_array)

        pointing["raw"]["eb_dcx"] = eb_dcx
        pointing["raw"]["eb_dcy"] = eb_dcy
        pointing["electrical"] = kmz_utils.acf_to_ecef(eb_dcx, eb_dcy, uacx, uacy)

        return pointing
    except Exception as exc:
        logger.warning(f"Exception while calculating {which} boresight")
        logger.warning(exc)
    return {}




from ..tre_elements import TREExtension, TREElement

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


class RSMPIAType(TREElement):
    def __init__(self, value):
        super(RSMPIAType, self).__init__()
        self.add_field('IID', 's', 80, value)
        self.add_field('EDITION', 's', 40, value)
        self.add_field('R0', 's', 21, value)
        self.add_field('RX', 's', 21, value)
        self.add_field('RY', 's', 21, value)
        self.add_field('RZ', 's', 21, value)
        self.add_field('RXX', 's', 21, value)
        self.add_field('RXY', 's', 21, value)
        self.add_field('RXZ', 's', 21, value)
        self.add_field('RYY', 's', 21, value)
        self.add_field('RYZ', 's', 21, value)
        self.add_field('RZZ', 's', 21, value)
        self.add_field('C0', 's', 21, value)
        self.add_field('CX', 's', 21, value)
        self.add_field('CY', 's', 21, value)
        self.add_field('CZ', 's', 21, value)
        self.add_field('CXX', 's', 21, value)
        self.add_field('CXY', 's', 21, value)
        self.add_field('CXZ', 's', 21, value)
        self.add_field('CYY', 's', 21, value)
        self.add_field('CYZ', 's', 21, value)
        self.add_field('CZZ', 's', 21, value)
        self.add_field('RNIS', 's', 3, value)
        self.add_field('CNIS', 's', 3, value)
        self.add_field('TNIS', 's', 3, value)
        self.add_field('RSSIZ', 's', 21, value)
        self.add_field('CSSIZ', 's', 21, value)


class RSMPIA(TREExtension):
    _tag_value = 'RSMPIA'
    _data_type = RSMPIAType



"""
Functions to map between the coordinates in image pixel space and geographical
coordinates.

Examples
--------

.. code-block:: python

    from sarpy.geometry import point_projection
    from sarpy.io.complex.sicd import SICDReader

    reader = SICDReader('<path to sicd file>')
    structure = reader.sicd_meta
    # or reader.reader.get_sicds_as_tuple()[0]

    # you can also use a SIDD structure, obtained from a product type reader

    # assume that ecf_coords is some previously defined numpy array of
    # shape (..., 3), with final dimension [X, Y, Z]
    image_coords = point_projection.ground_to_image(ecf_coords, structure)
    # image_coords will be a numpy array of shape (..., 2),
    # with final dimension [row, column]

    # assume that geo_coords is some previously defined numpy array of
    # shape (..., 3), with final dimension [lat, lon, hae]
    image_coords = point_projection.ground_to_image_geo(geo_coords, structure)
    # image_coords will be a numpy array of shape (..., 2),
    # with final dimension [row, column]

    # assume that image_coords is some previously defined numpy array of
    # shape (..., 2) with final dimension [row, column]
    ecf_coords_fixed_hae = point_projection.image_to_ground(image_coords, structure, projection_type='HAE')
    ecf_coords_plane = point_projection.image_to_ground(image_coords, structure, projection_type='PLANE')
    geo_coords_fixed_hae = point_projection.image_to_ground_geo(image_coords, structure, projection_type='HAE')
    geo_coords_plane = point_projection.image_to_ground_geo(image_coords, structure, projection_type='PLANE')
    # these outputs will be numpy arrays of shape (..., 3)

    # alternatively, these are also methods of the sicd/sidd structure
    image_coords = structure.project_ground_to_image(ecf_coords)
    image_coords = structure.project_ground_to_image_geo(geo_coords)
    ecf_coords_fixed_hae = structure.project_image_to_ground(image_coords, projection_type='HAE')
    ecf_coords_plane = structure.project_image_to_ground(image_coords, projection_type='PLANE')
    geo_coords_fixed_hae = structure.project_image_to_ground_geo(image_coords, projection_type='HAE')
    geo_coords_plane = structure.project_image_to_ground_geo(image_coords, projection_type='PLANE')


.. Note::

    It should be explicitly pointed out that these methods are essentially all
    inexact iterative methods which depend on convergence parameters. Changing
    these parameters will yield different numerically similar, but different
    results.

    Under the right assumptions involving the `projection_type` parameter value
    and the correct structure of the physical coordinates array, then the methods
    :meth:`ground_to_image` and :meth:`image_to_ground` are **approximate**
    inverses of one another. Being iterative methods, they can not generally be
    made numerically exact inverses, but the tolerance can be adjusted to yield
    very small differences.


.. Note::

    Virtually any SIDD/SICD structure which follows the standard will have an
    appropriate metadata populated to permit these projection methods.
    **In the case that your structure does not have sufficient metadata populated**,
    as may happen during research experimentation, an exception will be raised
    with hopefully helpful details about what information is missing.

"""

__classification__ = "UNCLASSIFIED"
__author__ = ("Thomas McCullough", "Wade Schwartzkopf")


import logging
from typing import Tuple, Union, Callable, Optional, List
from types import MethodType  # for binding a method dynamically to a class

import numpy

from sarpy.geometry.geocoords import ecf_to_geodetic, geodetic_to_ecf, wgs_84_norm
from sarpy.io.complex.sicd_elements.blocks import Poly2DType, XYZPolyType
from sarpy.io.DEM.DEM import DEMInterpolator
from sarpy.io.DEM.DTED import DTEDList, DTEDInterpolator
from sarpy.io.DEM.geoid import GeoidHeight


logger = logging.getLogger(__name__)

_unhandled_text = 'Got unhandled type `{}`'
_unsupported_text = 'Got unsupported projection type `{}`'


#############
# COA Projection definition

def _validate_adj_param(
        value: Union[None, numpy.ndarray, list, tuple],
        name: str) -> numpy.ndarray:
    """
    Validate the aperture adjustment vector parameters.

    Parameters
    ----------
    value : None|numpy.ndarray|list|tuple
    name : str

    Returns
    -------
    numpy.ndarray
    """

    if value is None:
        value = numpy.array([0, 0, 0], dtype='float64')
    if not isinstance(value, numpy.ndarray):
        value = numpy.array(value, dtype='float64')
    if value.shape != (3,):
        raise ValueError('{} must have shape (3, ). Got {}'.format(name, value.shape))
    return value


def _ric_ecf_mat(
        rarp: numpy.ndarray,
        varp: numpy.ndarray,
        frame_type: str) -> numpy.ndarray:
    """
    Computes the ECF transformation matrix for RIC frame.

    Parameters
    ----------
    rarp : numpy.ndarray
    varp : numpy.ndarray
    frame_type : str
        the final three characters should be one of ['ECI', 'ECF']

    Returns
    -------
    numpy.ndarray
        the RIC transform matrix (array)
    """

    # Angular velocity of earth in radians/second, not including precession
    w = 7292115.1467E-11
    typ = frame_type.upper()[-3:]
    vi = varp if typ == 'ECF' else varp + numpy.cross([0, 0, w], rarp)

    r = rarp/numpy.linalg.norm(rarp)
    c = numpy.cross(r, vi)
    c /= numpy.linalg.norm(c)  # NB: perpendicular to r
    i = numpy.cross(c, r)
    # this is the cross of two perpendicular normal vectors, so normal
    return numpy.stack([r, i, c], axis=-1).astype('float64')


def _get_sicd_type_specific_projection(sicd) -> Callable:
    """
    Gets an intermediate method specific projection method with six required
    calling arguments (self, row_transform, col_transform, time_coa, arp_coa, varp_coa).

    Parameters
    ----------
    sicd : sarpy.io.complex.sicd_elements.SICD.SICDType

    Returns
    -------
    callable
    """

    def pfa_projection():
        SCP = sicd.GeoData.SCP.ECF.get_array()
        pfa = sicd.PFA
        polar_ang_poly = pfa.PolarAngPoly
        spatial_freq_sf_poly = pfa.SpatialFreqSFPoly
        polar_ang_poly_der = polar_ang_poly.derivative(der_order=1, return_poly=True)
        spatial_freq_sf_poly_der = spatial_freq_sf_poly.derivative(der_order=1, return_poly=True)

        polar_ang_poly_der = polar_ang_poly.derivative(der_order=1, return_poly=True)
        spatial_freq_sf_poly_der = spatial_freq_sf_poly.derivative(der_order=1, return_poly=True)

        # noinspection PyUnusedLocal, PyIncorrectDocstring
        def method_projection(instance, row_transform, col_transform, time_coa, arp_coa, varp_coa):
            """
            PFA specific intermediate projection.

            Parameters
            ----------
            row_transform : numpy.ndarray
            col_transform : numpy.ndarray
            time_coa : numpy.ndarray
            arp_coa : numpy.ndarray
            varp_coa : numpy.ndarray

            Returns
            -------
            Tuple[numpy.ndarray, numpy.ndarray]
            """

            ARP_minus_SCP = arp_coa - SCP
            rSCPTgtCoa = numpy.linalg.norm(ARP_minus_SCP, axis=-1)
            rDotSCPTgtCoa = numpy.sum(varp_coa * ARP_minus_SCP, axis=-1) / rSCPTgtCoa

            thetaTgtCoa = polar_ang_poly(time_coa)
            dThetaDtTgtCoa = polar_ang_poly_der(time_coa)
            # Compute polar aperture scale factor (KSF) and derivative wrt polar angle
            ksfTgtCoa = spatial_freq_sf_poly(thetaTgtCoa)
            dKsfDThetaTgtCoa = spatial_freq_sf_poly_der(thetaTgtCoa)
            # Compute spatial frequency domain phase slopes in Ka and Kc directions
            # NB: sign for the phase may be ignored as it is cancelled in a subsequent computation.
            dPhiDKaTgtCoa = row_transform * numpy.cos(thetaTgtCoa) + col_transform * numpy.sin(thetaTgtCoa)
            dPhiDKcTgtCoa = -row_transform * numpy.sin(thetaTgtCoa) + col_transform * numpy.cos(thetaTgtCoa)
            # Compute range relative to SCP
            deltaRTgtCoa = ksfTgtCoa * dPhiDKaTgtCoa
            # Compute derivative of range relative to SCP wrt polar angle.
            # Scale by derivative of polar angle wrt time.
            dDeltaRDThetaTgtCoa = dKsfDThetaTgtCoa * dPhiDKaTgtCoa + ksfTgtCoa * dPhiDKcTgtCoa
            deltaRDotTgtCoa = dDeltaRDThetaTgtCoa * dThetaDtTgtCoa
            return rSCPTgtCoa + deltaRTgtCoa, rDotSCPTgtCoa + deltaRDotTgtCoa

        return method_projection

    def rgazcomp_projection():
        SCP = sicd.GeoData.SCP.ECF.get_array()
        az_sf = sicd.RgAzComp.AzSF

        # noinspection PyUnusedLocal, PyIncorrectDocstring
        def method_projection(instance, row_transform, col_transform, time_coa, arp_coa, varp_coa):
            """
            RgAzComp specific intermediate projection.

            Parameters
            ----------
            row_transform : numpy.ndarray
            col_transform : numpy.ndarray
            time_coa : numpy.ndarray
            arp_coa : numpy.ndarray
            varp_coa : numpy.ndarray

            Returns
            -------
            Tuple[numpy.ndarray, numpy.ndarray]
            """

            ARP_minus_SCP = arp_coa - SCP
            rSCPTgtCoa = numpy.linalg.norm(ARP_minus_SCP, axis=-1)
            rDotSCPTgtCoa = numpy.sum(varp_coa*ARP_minus_SCP, axis=-1)/rSCPTgtCoa
            deltaRTgtCoa = row_transform
            deltaRDotTgtCoa = -numpy.linalg.norm(varp_coa, axis=-1)*az_sf*col_transform
            return rSCPTgtCoa + deltaRTgtCoa, rDotSCPTgtCoa + deltaRDotTgtCoa

        return method_projection

    def inca_projection():
        inca = sicd.RMA.INCA
        r_ca_scp = inca.R_CA_SCP
        time_ca_poly = inca.TimeCAPoly
        drate_sf_poly = inca.DRateSFPoly

        # noinspection PyUnusedLocal, PyIncorrectDocstring
        def method_projection(instance, row_transform, col_transform, time_coa, arp_coa, varp_coa):
            """
            INCA specific intermediate projection.

            Parameters
            ----------
            row_transform : numpy.ndarray
            col_transform : numpy.ndarray
            time_coa : numpy.ndarray
            arp_coa : numpy.ndarray
            varp_coa : numpy.ndarray

            Returns
            -------
            Tuple[numpy.ndarray, numpy.ndarray]
            """

            # compute range/time of the closest approach
            R_CA_TGT = r_ca_scp + row_transform  # Range at closest approach
            t_CA_TGT = time_ca_poly(col_transform)  # Time of the closest approach
            # Compute ARP velocity magnitude (actually squared, since that's how it's used) at t_CA_TGT
            # noinspection PyProtectedMember
            VEL2_CA_TGT = numpy.sum(instance._varp_poly(t_CA_TGT)**2, axis=-1)
            # Compute the Doppler Rate Scale Factor for image Grid location
            DRSF_TGT = drate_sf_poly(row_transform, col_transform)
            # Difference between COA time and CA time
            dt_COA_TGT = time_coa - t_CA_TGT
            r_tgt_coa = numpy.sqrt(R_CA_TGT*R_CA_TGT + DRSF_TGT*VEL2_CA_TGT*dt_COA_TGT*dt_COA_TGT)
            r_dot_tgt_coa = (DRSF_TGT/r_tgt_coa)*VEL2_CA_TGT*dt_COA_TGT
            return r_tgt_coa, r_dot_tgt_coa

        return method_projection

    def plane_projection():
        SCP = sicd.GeoData.SCP.ECF.get_array()
        uRow = sicd.Grid.Row.UVectECF.get_array()
        uCol = sicd.Grid.Col.UVectECF.get_array()

        # noinspection PyUnusedLocal, PyIncorrectDocstring
        def method_projection(instance, row_transform, col_transform, time_coa, arp_coa, varp_coa):
            """
            Plane specific intermediate projection.

            Parameters
            ----------
            row_transform : numpy.ndarray
            col_transform : numpy.ndarray
            time_coa : numpy.ndarray
            arp_coa : numpy.ndarray
            varp_coa : numpy.ndarray

            Returns
            -------
            Tuple[numpy.ndarray, numpy.ndarray]
            """

            ARP_minus_IPP = arp_coa - (SCP + numpy.outer(row_transform, uRow) + numpy.outer(col_transform, uCol))
            r_tgt_coa = numpy.linalg.norm(ARP_minus_IPP, axis=-1)
            r_dot_tgt_coa = numpy.sum(varp_coa * ARP_minus_IPP, axis=-1)/r_tgt_coa
            return r_tgt_coa, r_dot_tgt_coa
        return method_projection

    # NB: sicd.can_project_coordinates() has been called, so all required attributes
    #   must be populated
    if sicd.Grid.Type == 'RGAZIM':
        if sicd.ImageFormation.ImageFormAlgo == 'PFA':
            return pfa_projection()
        elif sicd.ImageFormation.ImageFormAlgo == 'RGAZCOMP':
            return rgazcomp_projection()
    elif sicd.Grid.Type == 'RGZERO':
        return inca_projection()
    elif sicd.Grid.Type in ['XRGYCR', 'XCTYAT', 'PLANE']:
        return plane_projection()
    else:
        # NB: this will have been noted by sicd.can_project_coordinates(), but is
        #   here for completeness
        raise ValueError('Unhandled Grid.Type `{}`'.format(sicd.Grid.Type))


def _get_sicd_adjustment_params(
        sicd,
        delta_arp: Union[None, numpy.ndarray, list, tuple],
        delta_varp: Union[None, numpy.ndarray, list, tuple],
        adj_params_frame: str) -> Tuple[numpy.ndarray, numpy.ndarray]:
    """
    Gets the SICD adjustment params.

    Parameters
    ----------
    sicd : sarpy.io.complex.sicd_elements.SICD.SICDType
    delta_arp : None|numpy.ndarray|list|tuple
    delta_varp : None|numpy.ndarray|list|tuple
    adj_params_frame : str

    Returns
    -------
    delta_arp: numpy.ndarray
    delta_varp: numpy.ndarray
    """

    delta_arp = _validate_adj_param(delta_arp, 'delta_arp')
    delta_varp = _validate_adj_param(delta_varp, 'delta_varp')
    if adj_params_frame in ['RIC_ECI', 'RIC_ECF']:
        if sicd.SCPCOA.ARPPos is None or sicd.SCPCOA.ARPVel is None:
            raise ValueError(
                'The adj_params_frame is of RIC type, but one of SCPCOA.ARPPos or '
                'SCPCOA.ARPVel is not populated.')
        ARP_SCP_COA = sicd.SCPCOA.ARPPos.get_array()
        VARP_SCP_COA = sicd.SCPCOA.ARPVel.get_array()
        ric_matrix = _ric_ecf_mat(ARP_SCP_COA, VARP_SCP_COA, adj_params_frame)
        delta_arp = ric_matrix.dot(delta_arp)
        delta_varp = ric_matrix.dot(delta_varp)
    return delta_arp, delta_varp


def _get_sidd_type_projection(sidd) -> Union[Poly2DType, Callable]:
    """
    Gets an intermediate method specific projection method with six required
    calling arguments (self, row_transform, col_transform, time_coa, arp_coa, varp_coa).

    Parameters
    ----------
    sidd : SIDDType

    Returns
    -------
    (Poly2DType, callable)
    """

    def pgp(the_sidd):
        """

        Parameters
        ----------
        the_sidd : SIDDType

        Returns
        -------
        callable
        """
        plane_proj = the_sidd.Measurement.PlaneProjection

        SRP = plane_proj.ReferencePoint.ECEF.get_array()
        SRP_row = plane_proj.ReferencePoint.Point.Row
        SRP_col = plane_proj.ReferencePoint.Point.Col
        row_vector = plane_proj.ProductPlane.RowUnitVector.get_array()*plane_proj.SampleSpacing.Row
        col_vector = plane_proj.ProductPlane.ColUnitVector.get_array()*plane_proj.SampleSpacing.Col

        # noinspection PyUnusedLocal, PyIncorrectDocstring
        def method_projection(instance, row_transform, col_transform, time_coa, arp_coa, varp_coa):
            """
            Plane specific intermediate projection.

            Parameters
            ----------
            row_transform : numpy.ndarray
            col_transform : numpy.ndarray
            time_coa : numpy.ndarray
            arp_coa : numpy.ndarray
            varp_coa : numpy.ndarray

            Returns
            -------
            Tuple[numpy.ndarray, numpy.ndarray]
            """

            ARP_minus_IPP = arp_coa - \
                            (SRP + numpy.outer(row_transform - SRP_row, row_vector) +
                             numpy.outer(col_transform - SRP_col, col_vector))
            r_tgt_coa = numpy.linalg.norm(ARP_minus_IPP, axis=-1)
            r_dot_tgt_coa = numpy.sum(varp_coa * ARP_minus_IPP, axis=-1)/r_tgt_coa
            return r_tgt_coa, r_dot_tgt_coa
        return plane_proj.TimeCOAPoly, method_projection

    if sidd.Measurement.PlaneProjection is not None:
        return pgp(sidd)
    raise ValueError('Currently the only supported projection is PlaneProjection.')


def _get_sidd_adjustment_params(
        sidd,
        delta_arp: Union[None, numpy.ndarray, list, tuple],
        delta_varp: Union[None, numpy.ndarray, list, tuple],
        adj_params_frame: str) -> Tuple[numpy.ndarray, numpy.ndarray]:
    """
    Get the SIDD adjustment parameters.

    Parameters
    ----------
    sidd : SIDDType
    delta_arp : None|numpy.ndarray|list|tuple
    delta_varp : None|numpy.ndarray|list|tuple
    adj_params_frame : str

    Returns
    -------
    delta_arp: numpy.ndarray
    delta_varp: numpy.ndarray
    """

    delta_arp = _validate_adj_param(delta_arp, 'delta_arp')
    delta_varp = _validate_adj_param(delta_varp, 'delta_varp')

    if adj_params_frame in ['RIC_ECI', 'RIC_ECF']:
        arp_pos_poly = sidd.Measurement.ARPPoly
        arp_vel_poly = arp_pos_poly.derivative(der_order=1, return_poly=True)
        if sidd.Measurement.PlaneProjection is not None:
            srp_row = sidd.Measurement.PlaneProjection.ReferencePoint.Point.Row
            srp_col = sidd.Measurement.PlaneProjection.ReferencePoint.Point.Col
            srp_coa_time = sidd.Measurement.PlaneProjection.TimeCOAPoly(srp_row, srp_col)
            srp_pos = arp_pos_poly(srp_coa_time)
            srp_vel = arp_vel_poly(srp_coa_time)
            ric_matrix = _ric_ecf_mat(srp_pos, srp_vel, adj_params_frame)
            delta_arp = ric_matrix.dot(delta_arp)
            delta_varp = ric_matrix.dot(delta_varp)
        else:
            raise ValueError('Got unhandled projection type {}'.format(sidd.Measurement.ProjectionType))
    return delta_arp, delta_varp


class COAProjection(object):
    """
    The Center of Aperture projection object, which provides common projection
    functionality for all image to R/Rdot projection. This is a helper class,
    and generally not intended for direct usage.
    """

    __slots__ = (
        '_time_coa_poly', '_arp_poly', '_varp_poly', '_method_proj',
        '_row_shift', '_row_mult', '_col_shift', '_col_mult',
        '_delta_arp', '_delta_varp', '_range_bias',)

    def __init__(
            self,
            time_coa_poly: Poly2DType,
            arp_poly: XYZPolyType,
            method_projection: Callable,
            row_shift: Union[int, float] = 0,
            row_mult: Union[int, float] = 1,
            col_shift: Union[int, float] = 0,
            col_mult: Union[int, float] = 1,
            delta_arp: Union[None, numpy.ndarray, list, tuple] = None,
            delta_varp: Union[None, numpy.ndarray, list, tuple] = None,
            range_bias: Optional[float] = None):
        """

        Parameters
        ----------
        time_coa_poly : Poly2DType
            The time center of aperture polynomial.
        arp_poly : XYZPolyType
            The aperture position polynomial.
        method_projection : callable
            The method specific projection for performing the projection from image
            coordinates to R/Rdot space. The call signature is expected to be
            `method_projection(instance, row_transform, col_transform, time_coa, arp_coa, varp_coa)`,
            where `row_transform = row_mult*(row - row_shift)`,
            `col_transform = col_mult*(col - col_shift)`,
            `time_coa = time_coa_poly(row_transform, col_transform)`,
            `arp_coa = arp_poly(time_coa)`, and `varp_coa = varp_poly(time_coa)`.
        row_shift : int|float
            The shift part of the affine row transformation for plugging into the
            time coa polynomial.
        row_mult : int|float
            The multiple part of the affine row transformation for plugging into
            the time coa polynomial.
        col_shift : int|float
            The shift part of the affine column transformation for plugging into
            the time coa polynomial.
        col_mult : int|float
            The multiple part of the affine column transformation for plugging into
            the time coa polynomial.
        delta_arp : None|numpy.ndarray|list|tuple
            ARP position adjustable parameter (ECF, m).  Defaults to 0 in each coordinate.
        delta_varp : None|numpy.ndarray|list|tuple
            VARP position adjustable parameter (ECF, m/s).  Defaults to 0 in each coordinate.
        range_bias : float|int
            Range bias adjustable parameter (m), defaults to 0.
        """

        if not isinstance(time_coa_poly, Poly2DType):
            raise TypeError('time_coa_poly must be a Poly2DType instance.')
        self._time_coa_poly = time_coa_poly

        if not isinstance(arp_poly, XYZPolyType):
            raise TypeError('arp_poly must be an XYZPolyType instance.')
        self._arp_poly = arp_poly
        self._varp_poly = self._arp_poly.derivative(der_order=1, return_poly=True)  # type: XYZPolyType

        if not callable(method_projection):
            raise TypeError('method_projection must be callable.')
        self._method_proj = MethodType(method_projection, self)
        # affine transform parameters
        self._row_shift = float(row_shift)
        self._row_mult = float(row_mult)
        self._col_shift = float(col_shift)
        self._col_mult = float(col_mult)
        # aperture location adjustment parameters
        self._delta_arp = _validate_adj_param(delta_arp, 'delta_arp')
        self._delta_varp = _validate_adj_param(delta_varp, 'delta_varp')
        self._range_bias = 0.0 if range_bias is None else float(range_bias)  # type: float

    @property
    def delta_arp(self) -> numpy.ndarray:
        """
        numpy.ndarray: The delta arp adjustable parameter
        """

        return self._delta_arp

    @property
    def delta_varp(self) -> numpy.ndarray:
        """
        numpy.ndarray: The delta varp adjustable parameter
        """

        return self._delta_varp

    @property
    def range_bias(self) -> float:
        """
        float: The range bias adjustable parameter
        """

        return self._range_bias

    @property
    def delta_range(self) -> float:
        """
        float: Alias to the range bias adjustable parameter
        """

        return self._range_bias

    @classmethod
    def from_sicd(
            cls,
            sicd,
            delta_arp: Union[None, numpy.ndarray, list, tuple] = None,
            delta_varp: Union[None, numpy.ndarray, list, tuple] = None,
            range_bias: Optional[float] = None,
            adj_params_frame: str = 'ECF'):
        """
        Construct from a SICD structure.

        Parameters
        ----------
        sicd : sarpy.io.complex.sicd_elements.SICD.SICDType
            The SICD metadata structure.
        delta_arp : None|numpy.ndarray|list|tuple
            ARP position adjustable parameter (ECF, m).  Defaults to 0 in each coordinate.
        delta_varp : None|numpy.ndarray|list|tuple
            VARP position adjustable parameter (ECF, m/s).  Defaults to 0 in each coordinate.
        range_bias : float|int
            Range bias adjustable parameter (m), defaults to 0.
        adj_params_frame : str
            One of `('ECF', 'RIC_ECI', 'RIC_ECF')`.

        Returns
        -------
        COAProjection
        """

        if not sicd.can_project_coordinates():
            raise ValueError('Insufficient metadata populated to formulate projection.')

        time_coa_poly = sicd.Grid.TimeCOAPoly
        # fall back to approximation if TimeCOAPoly is not populated
        if time_coa_poly is None:
            time_coa_poly = Poly2DType(Coefs=[[sicd.Timeline.CollectDuration/2, ], ])
            logger.warning(
                'Using (constant) approximation to TimeCOAPoly, which may result in poor projection results.')
        arp_poly = sicd.Position.ARPPoly
        # transform parameters
        row_mult = sicd.Grid.Row.SS
        row_shift = sicd.ImageData.SCPPixel.Row - sicd.ImageData.FirstRow
        col_mult = sicd.Grid.Col.SS
        col_shift = sicd.ImageData.SCPPixel.Col - sicd.ImageData.FirstCol
        # location adjustment parameters
        delta_arp, delta_varp = _get_sicd_adjustment_params(sicd, delta_arp, delta_varp, adj_params_frame)
        return cls(time_coa_poly, arp_poly, _get_sicd_type_specific_projection(sicd),
                   row_shift=row_shift, row_mult=row_mult, col_shift=col_shift, col_mult=col_mult,
                   delta_arp=delta_arp, delta_varp=delta_varp, range_bias=range_bias)

    @classmethod
    def from_sidd(
            cls,
            sidd,
            delta_arp: Union[None, numpy.ndarray, list, tuple] = None,
            delta_varp: Union[None, numpy.ndarray, list, tuple] = None,
            range_bias: Optional[float] = None,
            adj_params_frame: str = 'ECF'):
        """
        Construct from the SIDD structure.

        Parameters
        ----------
        sidd : SIDDType
        delta_arp : None|numpy.ndarray|list|tuple
            ARP position adjustable parameter (ECF, m).  Defaults to 0 in each coordinate.
        delta_varp : None|numpy.ndarray|list|tuple
            VARP position adjustable parameter (ECF, m/s).  Defaults to 0 in each coordinate.
        range_bias : float|int
            Range bias adjustable parameter (m), defaults to 0.
        adj_params_frame : str
            One of `('ECF', 'RIC_ECI', 'RIC_ECF')`.

        Returns
        -------
        COAProjection
        """

        time_coa_poly, method_projection = _get_sidd_type_projection(sidd)
        arp_poly = sidd.Measurement.ARPPoly
        delta_arp, delta_varp = _get_sidd_adjustment_params(
            sidd, delta_arp, delta_varp, adj_params_frame)
        return cls(time_coa_poly, arp_poly, method_projection,
                   row_shift=0, row_mult=1, col_shift=0, col_mult=1,
                   delta_arp=delta_arp, delta_varp=delta_varp, range_bias=range_bias)

    def _init_proj(
            self,
            im_points: numpy.ndarray) -> Tuple[
                numpy.ndarray, numpy.ndarray, numpy.ndarray, numpy.ndarray, numpy.ndarray]:
        """

        Parameters
        ----------
        im_points : numpy.ndarray

        Returns
        -------
        row_transform: numpy.ndarray
        col_transform: numpy.ndarray
        time_coa: numpy.ndarray
        arp_coa: numpy.ndarray
        varp_coa: numpy.ndarray
        """

        row_transform = (im_points[:, 0] - self._row_shift)*self._row_mult
        col_transform = (im_points[:, 1] - self._col_shift)*self._col_mult
        time_coa = self._time_coa_poly(row_transform, col_transform)
        # calculate aperture reference position and velocity at target time
        arp_coa = self._arp_poly(time_coa)
        varp_coa = self._varp_poly(time_coa)
        return row_transform, col_transform, time_coa, arp_coa, varp_coa

    def projection(
            self,
            im_points: numpy.ndarray) -> Tuple[
                numpy.ndarray, numpy.ndarray, numpy.ndarray, numpy.ndarray, numpy.ndarray]:
        """
        Perform the projection from image coordinates to R/Rdot coordinates.

        Parameters
        ----------
        im_points : numpy.ndarray
            This array of image point coordinates, **expected to have shape (N, 2)**.

        Returns
        -------
        r_tgt_coa: numpy.ndarray
            range to the ARP at COA
        r_dot_tgt_coa: numpy.ndarray
            range rate relative to the ARP at COA
        time_coa: numpy.ndarray
            center of aperture time since CDP start for input ip
        arp_coa: numpy.ndarray
            aperture reference position at time_coa
        varp_coa: numpy.ndarray
            velocity at time_coa
        """

        row_transform, col_transform, time_coa, arp_coa, varp_coa = self._init_proj(im_points)
        r_tgt_coa, r_dot_tgt_coa = self._method_proj(row_transform, col_transform, time_coa, arp_coa, varp_coa)
        # adjust parameters
        arp_coa += self._delta_arp
        varp_coa += self._delta_varp
        r_tgt_coa += self._range_bias
        return r_tgt_coa, r_dot_tgt_coa, time_coa, arp_coa, varp_coa


def _get_coa_projection(
        structure,
        use_structure_coa: bool,
        **coa_args) -> COAProjection:
    """

    Parameters
    ----------
    structure
    use_structure_coa : bool
    coa_args

    Returns
    -------
    COAProjection
    """

    from sarpy.io.complex.sicd_elements.SICD import SICDType
    from sarpy.io.product.sidd import SIDD_TYPES

    if use_structure_coa and structure.coa_projection is not None:
        return structure.coa_projection
    elif isinstance(structure, SICDType):
        return COAProjection.from_sicd(structure, **coa_args)
    elif isinstance(structure, SIDD_TYPES):
        return COAProjection.from_sidd(structure, **coa_args)
    else:
        raise ValueError(_unhandled_text.format(type(structure)))


###############
# General helper methods for extracting params from the sicd or sidd

def _get_reference_point(structure) -> numpy.ndarray:
    """
    Gets the reference point in ECF coordinates.

    Parameters
    ----------
    structure

    Returns
    -------
    numpy.ndarray
    """

    from sarpy.io.complex.sicd_elements.SICD import SICDType
    from sarpy.io.product.sidd import SIDD_TYPES

    if isinstance(structure, SICDType):
        return structure.GeoData.SCP.ECF.get_array(dtype='float64')
    if isinstance(structure, SIDD_TYPES):
        proj_type = structure.Measurement.ProjectionType
        if proj_type != 'PlaneProjection':
            raise ValueError(_unsupported_text.format(proj_type))
        return structure.Measurement.PlaneProjection.ReferencePoint.ECEF.get_array(dtype='float64')
    raise TypeError(_unhandled_text.format(type(structure)))


def _get_outward_norm(structure, gref: numpy.ndarray) -> numpy.ndarray:
    """
    Gets the default outward unit norm.

    Parameters
    ----------
    structure
    gref : numpy.ndarray

    Returns
    -------
    numpy.ndarray
    """

    from sarpy.io.complex.sicd_elements.SICD import SICDType
    from sarpy.io.product.sidd import SIDD_TYPES

    if isinstance(structure, SICDType):
        if structure.ImageFormation.ImageFormAlgo == 'PFA':
            return structure.PFA.FPN.get_array()
        return wgs_84_norm(gref)
    if isinstance(structure, SIDD_TYPES):
        proj_type = structure.Measurement.ProjectionType
        if proj_type != 'PlaneProjection':
            raise ValueError(_unsupported_text.format(proj_type))
        the_proj = structure.Measurement.PlaneProjection
        # image plane details
        uRow = the_proj.ProductPlane.RowUnitVector.get_array(dtype='float64')
        uCol = the_proj.ProductPlane.ColUnitVector.get_array(dtype='float64')
        # outward unit norm for plane
        uGPN = numpy.cross(uRow, uCol)
        uGPN /= numpy.linalg.norm(uGPN)
        if numpy.dot(uGPN, gref) < 0:
            uGPN *= -1
        return uGPN
    raise TypeError(_unhandled_text.format(type(structure)))


def _extract_plane_params(structure) -> Tuple[
        numpy.ndarray, numpy.ndarray, float, float, numpy.ndarray, numpy.ndarray, numpy.ndarray, numpy.ndarray]:
    """
    Extract the required parameters for projection from ground to plane for a SICD.

    Parameters
    ----------
    structure : sarpy.io.complex.sicd_elements.SICD.SICDType|sarpy.io.product.sidd2_elements.SIDD.SIDDType|sarpy.io.product.sidd1_elements.SIDD.SIDDType

    Returns
    -------
    ref_point: numpy.ndarray
    ref_pixel: numpy.ndarray
    row_ss: float
    col_ss: float
    uRow: numpy.ndarray
    uCol: numpy.ndarray
    uGPN: numpy.ndarray
    uSPN: numpy.ndarray
    """

    from sarpy.io.complex.sicd_elements.SICD import SICDType
    from sarpy.io.product.sidd import SIDD_TYPES

    if isinstance(structure, SICDType):
        # reference point for the plane
        ref_point = structure.GeoData.SCP.ECF.get_array()
        ref_pixel = structure.ImageData.SCPPixel.get_array()

        # pixel spacing
        row_ss = structure.Grid.Row.SS
        col_ss = structure.Grid.Col.SS

        # image plane details
        uRow = structure.Grid.Row.UVectECF.get_array()  # unit normal in row direction
        uCol = structure.Grid.Col.UVectECF.get_array()  # unit normal in column direction

        # outward unit norm
        uGPN = structure.PFA.FPN.get_array() if structure.ImageFormation.ImageFormAlgo == 'PFA' \
            else wgs_84_norm(ref_point)

        # uSPN - defined in section 3.1 as normal to instantaneous slant plane that contains SCP at SCP COA is
        # tangent to R/Rdot contour at SCP. Points away from center of Earth. Use look to establish sign.
        ARP_SCP_COA = structure.SCPCOA.ARPPos.get_array()
        VARP_SCP_COA = structure.SCPCOA.ARPVel.get_array()
        uSPN = structure.SCPCOA.look*numpy.cross(VARP_SCP_COA, ref_point - ARP_SCP_COA)
        uSPN /= numpy.linalg.norm(uSPN)

        return ref_point, ref_pixel, row_ss, col_ss, uRow, uCol, uGPN, uSPN
    if isinstance(structure, SIDD_TYPES):
        proj_type = structure.Measurement.ProjectionType
        if proj_type != 'PlaneProjection':
            raise ValueError(_unsupported_text.format(proj_type))

        the_proj = structure.Measurement.PlaneProjection

        # reference point for the plane
        ref_point = the_proj.ReferencePoint.ECEF.get_array(dtype='float64')
        ref_pixel = the_proj.ReferencePoint.Point.get_array(dtype='float64')
        # pixel spacing
        row_ss = the_proj.SampleSpacing.Row
        col_ss = the_proj.SampleSpacing.Col
        # image plane details
        uRow = the_proj.ProductPlane.RowUnitVector.get_array(dtype='float64')
        uCol = the_proj.ProductPlane.ColUnitVector.get_array(dtype='float64')
        # outward unit norm for plane
        uGPN = numpy.cross(uRow, uCol)
        uGPN /= numpy.linalg.norm(uGPN)
        if numpy.dot(uGPN, ref_point) < 0:
            uGPN *= -1

        # slant plane is identical to outward unit norm
        return ref_point, ref_pixel, row_ss, col_ss, uRow, uCol, uGPN, uGPN
    raise TypeError('Got structure unsupported type {}'.format(type(structure)))


#############
# Ground-to-Image (aka Scene-to-Image) projection.

def _validate_coords(coords: numpy.ndarray) -> Tuple[numpy.ndarray, Tuple[int, ...]]:
    if not isinstance(coords, numpy.ndarray):
        coords = numpy.array(coords, dtype='float64')

    orig_shape = coords.shape

    if len(orig_shape) == 1:
        coords = numpy.reshape(coords, (1, -1))
    if coords.shape[-1] != 3:
        raise ValueError(
            'The coords array must represent an array of points in ECF coordinates, '
            'so the final dimension of coords must have length 3. Have coords.shape = {}'.format(coords.shape))

    return coords, orig_shape


def _ground_to_image(
        coords: numpy.ndarray,
        coa_proj: COAProjection,
        uGPN: numpy.ndarray,
        ref_point: numpy.ndarray,
        ref_pixel: numpy.ndarray,
        uIPN: numpy.ndarray,
        sf: float,
        row_ss: float,
        col_ss: float,
        uProj: numpy.ndarray,
        row_col_transform: numpy.ndarray,
        ipp_transform: numpy.ndarray,
        tolerance: float,
        max_iterations: int) -> Tuple[numpy.ndarray, numpy.ndarray, int]:
    """
    Basic level helper function.

    Parameters
    ----------
    coords : numpy.ndarray|tuple|list
    coa_proj : COAProjection
    uGPN : numpy.ndarray
    ref_point : numpy.ndarray
    ref_pixel : numpy.ndarray
    uIPN : numpy.ndarray
    sf : float
    row_ss : float
    col_ss : float
    uProj : numpy.ndarray
    row_col_transform : numpy.ndarray
    ipp_transform : numpy.ndarray
    tolerance : float
    max_iterations : int

    Returns
    -------
    image_points: numpy.ndarray
        The determined image point array, of size `N x 2`. Following SICD
        convention, the upper-left pixel is [0, 0].
    delta_gpn: numpy.ndarray
        Residual ground plane displacement (m).
    iterations: int
        The number of iterations performed.
    """

    g_n = coords.copy()
    im_points = numpy.zeros((coords.shape[0], 2), dtype='float64')
    delta_gpn = numpy.zeros((coords.shape[0],), dtype='float64')
    cont = True
    iteration = 0

    matrix_transform = numpy.dot(row_col_transform, ipp_transform)
    # (3 x 2)*(2 x 2) = (3 x 2)

    while cont:
        # project ground plane to image plane iteration
        iteration += 1
        dist_n = numpy.dot(ref_point - g_n, uIPN)/sf  # (N, )
        i_n = g_n + numpy.outer(dist_n, uProj)  # (N, 3)
        delta_ipp = i_n - ref_point  # (N, 3)
        ip_iter = numpy.dot(delta_ipp, matrix_transform)  # (N, 2)
        im_points[:, 0] = ip_iter[:, 0]/row_ss + ref_pixel[0]
        im_points[:, 1] = ip_iter[:, 1]/col_ss + ref_pixel[1]
        # transform to ground plane containing the scene points and check how it compares
        p_n = _image_to_ground_plane(im_points, coa_proj, g_n, uGPN)
        # compute displacement between scene point and this new projected point
        diff_n = coords - p_n
        delta_gpn[:] = numpy.linalg.norm(diff_n, axis=1)
        g_n += diff_n

        # should we continue iterating?
        cont = numpy.any(delta_gpn > tolerance) and (iteration < max_iterations)

    return im_points, delta_gpn, iteration


def ground_to_image(
        coords: Union[numpy.ndarray, list, tuple],
        structure,
        tolerance: float = 1e-2,
        max_iterations: int = 10,
        block_size: Optional[int] = 50000,
        use_structure_coa: bool = True,
        **coa_args) -> Tuple[numpy.ndarray, Union[numpy.ndarray, float], Union[numpy.ndarray, int]]:
    """
    Transforms a 3D ECF point to pixel (row/column) coordinates. This is
    implemented in accordance with the SICD Image Projections Description Document.
    **Really Scene-To-Image projection.**"

    Parameters
    ----------
    coords : numpy.ndarray|tuple|list
        ECF coordinate to map to scene coordinates, of size `N x 3`.
    structure : sarpy.io.complex.sicd_elements.SICD.SICDType|sarpy.io.product.sidd2_elements.SIDD.SIDDType|sarpy.io.product.sidd1_elements.SIDD.SIDDType
        The SICD or SIDD data structure.
    tolerance : float|int
        Ground plane displacement tol (m).
    max_iterations : int
        maximum number of iterations to perform
    block_size : int|None
        size of blocks of coordinates to transform at a time
    use_structure_coa : bool
        If sicd.coa_projection is populated, use that one **ignoring the COAProjection parameters.**
    coa_args
        The keyword arguments from the COAProjection.from_sicd class method.

    Returns
    -------
    image_points: numpy.ndarray
        The determined image point array. Following the SICD convention, t
        the upper-left pixel is [0, 0].
    delta_gpn: numpy.ndarray|float
        The residual ground plane displacement (m).
    iterations: numpy.ndarray|int
        The number of iterations performed.
    """

    coords, orig_shape = _validate_coords(coords)
    coa_proj = _get_coa_projection(structure, use_structure_coa, **coa_args)

    ref_point, ref_pixel, row_ss, col_ss, uRow, uCol, \
    uGPN, uSPN = _extract_plane_params(structure)

    uIPN = numpy.cross(uRow, uCol)  # NB: only outward pointing if Row/Col are right-handed system
    uIPN /= numpy.linalg.norm(uIPN)  # NB: uRow/uCol may not be perpendicular

    cos_theta = numpy.dot(uRow, uCol)
    sin_theta = numpy.sqrt(1 - cos_theta*cos_theta)
    ipp_transform = numpy.array(
        [[1, -cos_theta], [-cos_theta, 1]], dtype='float64')/(sin_theta*sin_theta)
    row_col_transform = numpy.zeros((3, 2), dtype='float64')
    row_col_transform[:, 0] = uRow
    row_col_transform[:, 1] = uCol
    sf = float(numpy.dot(uSPN, uIPN))  # scale factor

    tolerance = float(tolerance)
    if tolerance < 1e-12:
        logger.warning(
            'minimum allowed tolerance is 1e-12 meters, resetting from {}'.format(tolerance))
        tolerance = 1e-12

    # prepare the work space
    coords_view = numpy.reshape(coords, (-1, 3))  # possibly or make 2-d flatten
    num_points = coords_view.shape[0]
    if block_size is None or num_points <= block_size:
        image_points, delta_gpn, iters = _ground_to_image(
            coords_view, coa_proj, uGPN,
            ref_point, ref_pixel, uIPN, sf, row_ss, col_ss, uSPN,
            row_col_transform, ipp_transform, tolerance, max_iterations)
        iters = numpy.full((num_points, ), iters)
    else:
        image_points = numpy.zeros((num_points, 2), dtype='float64')
        delta_gpn = numpy.zeros((num_points, ), dtype='float64')
        iters = numpy.zeros((num_points, ), dtype='int16')

        # proceed with block processing
        start_block = 0
        while start_block < num_points:
            end_block = min(start_block+block_size, num_points)
            image_points[start_block:end_block, :], delta_gpn[start_block:end_block], \
                iters[start_block:end_block] = _ground_to_image(
                    coords_view[start_block:end_block, :], coa_proj, uGPN,
                    ref_point, ref_pixel, uIPN, sf, row_ss, col_ss, uSPN,
                    row_col_transform, ipp_transform, tolerance, max_iterations)
            start_block = end_block

    if len(orig_shape) == 1:
        image_points = numpy.reshape(image_points, (-1,))
        delta_gpn = float(delta_gpn[0])
        iters = int(iters[0])
    elif len(orig_shape) > 1:
        image_points = numpy.reshape(image_points, orig_shape[:-1]+(2, ))
        delta_gpn = numpy.reshape(delta_gpn, orig_shape[:-1])
        iters = numpy.reshape(iters, orig_shape[:-1])
    return image_points, delta_gpn, iters


def ground_to_image_geo(
        coords,
        structure,
        ordering='latlong',
        **kwargs) -> Tuple[numpy.ndarray, Union[numpy.ndarray, float], Union[numpy.ndarray, int]]:
    """
    Transforms a 3D Lat/Lon/HAE point to pixel (row/column) coordinates.
    This is implemented in accordance with the SICD Image Projections Description Document.

    Parameters
    ----------
    coords : numpy.ndarray|tuple|list
        Lat/Lon/HAE coordinate to map to scene coordinates, of size `N x 3`.
    structure : sarpy.io.complex.sicd_elements.SICD.SICDType|sarpy.io.product.sidd2_elements.SIDD.SIDDType|sarpy.io.product.sidd1_elements.SIDD.SIDDType
        The SICD or SIDD structure.
    ordering : str
        If 'longlat', then the input is `[longitude, latitude, hae]`.
        Otherwise, the input is `[latitude, longitude, hae]`. Passed through
        to :func:`sarpy.geometry.geocoords.geodetic_to_ecf`.
    kwargs
        See the key word arguments of :func:`ground_to_image`

    Returns
    -------
    image_points: numpy.ndarray
        The determined image point array. Following the SICD convention,
        the upper-left pixel is [0, 0].
    delta_gpn: numpy.ndarray|float
        The residual ground plane displacement (m).
    iterations: numpy.ndarray|int
        The number of iterations performed.
    """

    return ground_to_image(geodetic_to_ecf(coords, ordering=ordering), structure, **kwargs)


############
# Image-To-Ground projections

def _validate_im_points(
        im_points: Union[numpy.ndarray, list, tuple]) -> Tuple[numpy.ndarray, Tuple[int, ...]]:
    """

    Parameters
    ----------
    im_points : numpy.ndarray|list|tuple

    Returns
    -------
    im_points: numpy.ndarray
    orig_shape: Tuple[int, ...]
    """

    if im_points is None:
        raise ValueError('The argument cannot be None')

    if not isinstance(im_points, numpy.ndarray):
        im_points = numpy.array(im_points, dtype='float64')

    orig_shape = im_points.shape

    if len(im_points.shape) == 1:
        im_points = numpy.reshape(im_points, (1, -1))
    if im_points.shape[-1] != 2:
        raise ValueError(
            'The im_points array must represent an array of points in pixel coordinates, '
            'so the final dimension of im_points must have length 2. '
            'Have im_points.shape = {}'.format(im_points.shape))
    return im_points, orig_shape


def image_to_ground(
        im_points: Union[numpy.ndarray, list, tuple],
        structure,
        block_size: Optional[int] = 50000,
        projection_type: str = 'HAE',
        use_structure_coa: bool = True,
        **kwargs) -> numpy.ndarray:
    """
    Transforms image coordinates to ground plane ECF coordinate via the algorithm(s)
    described in SICD Image Projections document.

    Parameters
    ----------
    im_points : numpy.ndarray|list|tuple
        (row, column) coordinates of N points in image (or subimage if FirstRow/FirstCol are nonzero).
        Following SICD convention, the upper-left pixel is [0, 0].
    structure : SICDType|SIDDType
        The SICD or SIDD structure.
    block_size : None|int
        Size of blocks of coordinates to transform at a time. The entire array will be
        transformed as a single block if `None`.
    projection_type : str
        One of ['PLANE', 'HAE', 'DEM'].
    use_structure_coa : bool
        If structure.coa_projection is populated, use that one **ignoring the COAProjection parameters.**
    kwargs
        keyword arguments relevant for the given projection type. See image_to_ground_plane/hae/dem methods.

    Returns
    -------
    numpy.ndarray
        Physical coordinates (in ECF) corresponding input image coordinates. The interpretation
        or meaning of the physical coordinates depends on `projection_type` chosen.
    """

    p_type = projection_type.upper()
    if p_type == 'PLANE':
        return image_to_ground_plane(
            im_points, structure, block_size=block_size, use_structure_coa=use_structure_coa, **kwargs)
    elif p_type == 'HAE':
        return image_to_ground_hae(
            im_points, structure, block_size=block_size, use_structure_coa=use_structure_coa, **kwargs)
    elif p_type == 'DEM':
        return image_to_ground_dem(
            im_points, structure, block_size=block_size, use_structure_coa=use_structure_coa, **kwargs)
    else:
        raise ValueError('Got unrecognized projection type {}'.format(projection_type))


def image_to_ground_geo(
        im_points: Union[numpy.ndarray, list, tuple],
        structure,
        ordering: str = 'latlong',
        block_size: Optional[int] = 50000,
        projection_type: str = 'HAE',
        use_structure_coa: bool = True,
        **kwargs) -> numpy.ndarray:
    """
    Transforms image coordinates to ground plane Lat/Lon/HAE coordinate via the algorithm(s)
    described in SICD Image Projections document.

    Parameters
    ----------
    im_points : numpy.ndarray|list|tuple
        (row, column) coordinates of N points in image (or subimage if FirstRow/FirstCol are nonzero).
        Following SICD convention, the upper-left pixel is [0, 0].
    structure : SICDType|SIDDType
        The SICD or SIDD structure.
    ordering : str
        Determines whether return is ordered as `[lat, long, hae]` or `[long, lat, hae]`.
        Passed through to :func:`sarpy.geometry.geocoords.ecf_to_geodetic`.
    block_size : None|int
        Size of blocks of coordinates to transform at a time. The entire array will be
        transformed as a single block if `None`.
    projection_type : str
        One of ['PLANE', 'HAE', 'DEM'].
    use_structure_coa : bool
        If structure.coa_projection is populated, use that one **ignoring the COAProjection parameters.**
    kwargs
        See the keyword arguments in :func:`image_to_ground`.

    Returns
    -------
    numpy.ndarray
        Ground Plane Point (in Lat/Lon/HAE coordinates) along the R/Rdot contour.
    """

    return ecf_to_geodetic(
        image_to_ground(
            im_points, structure, block_size=block_size, projection_type=projection_type,
            use_structure_coa=use_structure_coa, **kwargs),
        ordering=ordering)


#####
# Image-to-Ground Plane

def _image_to_ground_plane_perform(
        r_tgt_coa: numpy.ndarray,
        r_dot_tgt_coa: numpy.ndarray,
        arp_coa: numpy.ndarray,
        varp_coa: numpy.ndarray,
        gref: numpy.ndarray,
        uZ: numpy.ndarray) -> numpy.ndarray:
    """

    Parameters
    ----------
    r_tgt_coa : numpy.ndarray
    r_dot_tgt_coa : numpy.ndarray
    arp_coa : numpy.ndarray
    varp_coa : numpy.ndarray
    gref : numpy.ndarray
    uZ : numpy.ndarray

    Returns
    -------
    numpy.ndarray
    """

    # Solve for the intersection of a R/Rdot contour and a ground plane.
    arpZ = numpy.sum((arp_coa - gref)*uZ, axis=-1)
    arpZ[arpZ > r_tgt_coa] = numpy.nan
    # ARP ground plane nadir
    aGPN = arp_coa - numpy.outer(arpZ, uZ)
    # Compute ground plane distance (gd) from ARP nadir to circle of const range
    gd = numpy.sqrt(r_tgt_coa*r_tgt_coa - arpZ*arpZ)
    # Compute sine and cosine of grazing angle
    cosGraz = gd/r_tgt_coa
    sinGraz = arpZ/r_tgt_coa

    # Velocity components normal to ground plane and parallel to ground plane.
    vMag = numpy.linalg.norm(varp_coa, axis=-1)
    vZ = numpy.dot(varp_coa, uZ)
    vX = numpy.sqrt(vMag*vMag - vZ*vZ)  # Note: For Vx = 0, no Solution
    # Orient X such that Vx > 0 and compute unit vectors uX and uY
    uX = (varp_coa - numpy.outer(vZ, uZ))/vX[:, numpy.newaxis]
    uY = numpy.cross(uZ, uX)
    # Compute cosine of azimuth angle to ground plane point
    cosAz = (-r_dot_tgt_coa+vZ*sinGraz) / (vX * cosGraz)
    cosAz[numpy.abs(cosAz) > 1] = numpy.nan  # R/Rdot combination not possible in given plane

    # Compute sine of azimuth angle. Use LOOK to establish sign.
    look = numpy.sign(numpy.dot(numpy.cross(arp_coa-gref, varp_coa), uZ))
    sinAz = look*numpy.sqrt(1-cosAz*cosAz)

    # Compute Ground Plane Point in ground plane and along the R/Rdot contour
    return aGPN + uX*(gd*cosAz)[:, numpy.newaxis] + uY*(gd*sinAz)[:, numpy.newaxis]


def _image_to_ground_plane(
        im_points: numpy.ndarray,
        coa_projection: COAProjection,
        gref: numpy.ndarray,
        uZ: numpy.ndarray) -> numpy.ndarray:
    """

    Parameters
    ----------
    im_points : numpy.ndarray
    coa_projection : COAProjection
    gref : numpy.ndarray
    uZ : numpy.ndarray

    Returns
    -------
    numpy.ndarray
    """

    r_tgt_coa, r_dot_tgt_coa, time_coa, arp_coa, varp_coa = coa_projection.projection(im_points)
    values = _image_to_ground_plane_perform(r_tgt_coa, r_dot_tgt_coa, arp_coa, varp_coa, gref, uZ)
    return values


def image_to_ground_plane(
        im_points: Union[numpy.ndarray, list, tuple],
        structure,
        block_size: Optional[int] = 50000,
        gref: Union[None, numpy.ndarray, list, tuple] = None,
        ugpn: Union[None, numpy.ndarray, list, tuple] = None,
        use_structure_coa: bool = True,
        **coa_args):
    """
    Transforms image coordinates to ground plane ECF coordinate via the algorithm(s)
    described in SICD Image Projections document.

    Parameters
    ----------
    im_points : numpy.ndarray|list|tuple
        the image coordinate array
    structure : SICDType|SIDDType
        The SICD or SIDD structure.
    block_size : None|int
        Size of blocks of coordinates to transform at a time. The entire array will be
        transformed as a single block if `None`.
    gref : None|numpy.ndarray|list|tuple
        Ground plane reference point ECF coordinates (m). The default is the SCP or Reference Point.
    ugpn : None|numpy.ndarray|list|tuple
        Vector normal to the plane to which we are projecting.
    use_structure_coa : bool
        If structure.coa_projection is populated, use that one **ignoring the COAProjection parameters.**
    coa_args
        keyword arguments for COAProjection.from_sicd class method.

    Returns
    -------
    numpy.ndarray
        Ground Plane Point (in ECF coordinates) corresponding to the input image coordinates.
    """

    im_points, orig_shape = _validate_im_points(im_points)
    coa_proj = _get_coa_projection(structure, use_structure_coa, **coa_args)

    # method parameter validation
    if gref is None:
        gref = _get_reference_point(structure)
    if not isinstance(gref, numpy.ndarray):
        gref = numpy.array(gref, dtype='float64')
    if gref.size != 3:
        raise ValueError('gref must have three elements.')
    if gref.ndim != 1:
        gref = numpy.reshape(gref, (3, ))

    if ugpn is None:
        ugpn = _get_outward_norm(structure, gref)
    if not isinstance(ugpn, numpy.ndarray):
        ugpn = numpy.array(ugpn, dtype='float64')
    if ugpn.size != 3:
        raise ValueError('ugpn must have three elements.')
    if ugpn.ndim != 1:
        ugpn = numpy.reshape(ugpn, (3, ))
    uZ = ugpn/numpy.linalg.norm(ugpn)

    # prepare workspace
    im_points_view = numpy.reshape(im_points, (-1, 2))  # possibly or make 2-d flatten
    num_points = im_points_view.shape[0]
    if block_size is None or num_points <= block_size:
        coords = _image_to_ground_plane(im_points_view, coa_proj, gref, uZ)
    else:
        coords = numpy.zeros((num_points, 3), dtype='float64')
        # proceed with block processing
        start_block = 0
        while start_block < num_points:
            end_block = min(start_block + block_size, num_points)
            coords[start_block:end_block, :] = _image_to_ground_plane(
                im_points_view[start_block:end_block], coa_proj, gref, uZ)
            start_block = end_block

    if len(orig_shape) == 1:
        coords = numpy.reshape(coords, (-1, ))
    elif len(orig_shape) > 1:
        coords = numpy.reshape(coords, orig_shape[:-1] + (3,))
    return coords


#####
# Image-to-HAE

def _image_to_ground_hae_perform(
        r_tgt_coa: numpy.ndarray,
        r_dot_tgt_coa: numpy.ndarray,
        arp_coa: numpy.ndarray,
        varp_coa: numpy.ndarray,
        ref_point: numpy.ndarray,
        ugpn: numpy.ndarray,
        hae0: float,
        tolerance: float,
        max_iterations: int,
        ref_hae: float) -> numpy.ndarray:
    """
    Intermediate helper method.

    Parameters
    ----------
    r_tgt_coa : numpy.ndarray
    r_dot_tgt_coa : numpy.ndarray
    arp_coa : numpy.ndarray
    varp_coa : numpy.ndarray
    ref_point : numpy.ndarray
    ugpn : numpy.ndarray
    hae0 : float
    tolerance : float
    max_iterations : int
    ref_hae : float

    Returns
    -------
    numpy.ndarray
    """

    # Compute the geodetic ground plane normal at the ref_point.
    look = numpy.sign(numpy.sum(numpy.cross(arp_coa, varp_coa)*(ref_point - arp_coa), axis=1))
    gref = ref_point - (ref_hae - hae0)*ugpn
    # iteration variables
    gpp = None
    delta_hae = None
    cont = True
    iters = 0
    while cont:
        iters += 1
        # Compute the precise projection along the R/Rdot contour to Ground Plane.
        gpp = _image_to_ground_plane_perform(r_tgt_coa, r_dot_tgt_coa, arp_coa, varp_coa, gref, ugpn)
        # check our hae value versus hae0
        gpp_llh = ecf_to_geodetic(gpp)
        delta_hae = gpp_llh[:, 2] - hae0
        max_abs_delta_hae = numpy.max(numpy.abs(delta_hae))
        gref = gpp - (delta_hae[:, numpy.newaxis] * ugpn)
        # should we stop our iteration?
        cont = (max_abs_delta_hae > tolerance) and (iters < max_iterations)

    # Compute the unit slant plane normal vector, uspn, that is tangent to the R/Rdot contour at point gpp
    uspn = numpy.cross(varp_coa, (gpp - arp_coa))*look[:, numpy.newaxis]
    uspn /= numpy.linalg.norm(uspn, axis=-1)[:, numpy.newaxis]
    # For the final straight line projection, project from point gpp along
    # the slant plane normal (as opposed to the ground plane normal that was
    # used in the iteration) to point slp.
    sf = numpy.sum(ugpn*uspn, axis=-1)
    slp = gpp - uspn*(delta_hae/sf)[:, numpy.newaxis]
    # Assign surface point SPP position by adjusting the HAE to be on the
    # HAE0 surface.
    spp_llh = ecf_to_geodetic(slp)
    spp_llh[:, 2] = hae0
    spp = geodetic_to_ecf(spp_llh)
    return spp


def _image_to_ground_hae(
        im_points: numpy.ndarray,
        coa_projection: COAProjection,
        hae0: float,
        tolerance: float,
        max_iterations: int,
        ref_hae: float,
        ref_point: numpy.ndarray) -> numpy.ndarray:
    """
    Intermediate helper function for projection.

    Parameters
    ----------
    im_points : numpy.ndarray
        the image coordinate array
    coa_projection : COAProjection
    hae0 : float
    tolerance : float
    max_iterations : int
    ref_hae : float
    ref_point : numpy.ndarray

    Returns
    -------
    numpy.ndarray
    """

    # get (image formation specific) projection parameters
    r_tgt_coa, r_dot_tgt_coa, time_coa, arp_coa, varp_coa = coa_projection.projection(im_points)
    ugpn = wgs_84_norm(ref_point)
    return _image_to_ground_hae_perform(
        r_tgt_coa, r_dot_tgt_coa, arp_coa, varp_coa, ref_point, ugpn,
        hae0, tolerance, max_iterations, ref_hae)


def image_to_ground_hae(
        im_points: Union[numpy.ndarray, list, tuple],
        structure,
        block_size: Optional[int] = 50000,
        hae0: Optional[float] = None,
        tolerance: float = 1e-3,
        max_iterations: int = 10,
        use_structure_coa: bool = True,
        **coa_args) -> numpy.ndarray:
    """
    Transforms image coordinates to ground plane ECF coordinate via the algorithm(s)
    described in SICD Image Projections document.

    Parameters
    ----------
    im_points : numpy.ndarray|list|tuple
        the image coordinate array
    structure : SICDType|SIDDType
        The SICD or SIDD structure.
    block_size : None|int
        Size of blocks of coordinates to transform at a time. The entire array will be
        transformed as a single block if `None`.
    hae0 : None|float|int
        Surface height (m) above the WGS-84 reference ellipsoid for projection point.
        Defaults to HAE at the SCP or Reference Point.
    tolerance : float|int
        Height threshold for convergence of iterative constant HAE computation (m).
    max_iterations : int
        Maximum number of iterations allowed for constant hae computation.
    use_structure_coa : bool
        If structure.coa_projection is populated, use that one **ignoring the COAProjection parameters.**
    coa_args
        keyword arguments for COAProjection.from_sicd class method.

    Returns
    -------
    numpy.ndarray
        Ground Plane Point (in ECF coordinates) with target hae corresponding to
        the input image coordinates.
    """

    # coa projection creation
    im_points, orig_shape = _validate_im_points(im_points)
    coa_proj = _get_coa_projection(structure, use_structure_coa, **coa_args)

    tolerance = float(tolerance)
    if tolerance < 1e-12:
        logger.warning(
            'minimum allowed tolerance is 1e-12, resetting from {0:8f}'.format(tolerance))
        tolerance = 1e-12

    max_iterations = int(max_iterations)
    if max_iterations < 1:
        logger.error(
            'max_iterations must be a positive integer, resetting to 1 from {}'.format(max_iterations))
        max_iterations = 1
    if max_iterations > 100:
        logger.error(
            'maximum allowed max_iterations is 100, resetting from {}'.format(max_iterations))
        max_iterations = 100

    # method parameter validation
    ref_point = _get_reference_point(structure)
    ref_llh = ecf_to_geodetic(ref_point)
    ref_hae = float(ref_llh[2])
    if hae0 is None:
        hae0 = ref_hae

    # prepare workspace
    im_points_view = numpy.reshape(im_points, (-1, 2))  # possibly or make 2-d flatten
    num_points = im_points_view.shape[0]
    if block_size is None or num_points <= block_size:
        coords = _image_to_ground_hae(im_points_view, coa_proj, hae0, tolerance, max_iterations, ref_hae, ref_point)
    else:
        coords = numpy.zeros((num_points, 3), dtype='float64')
        # proceed with block processing
        start_block = 0
        while start_block < num_points:
            end_block = min(start_block + block_size, num_points)
            coords[start_block:end_block, :] = _image_to_ground_hae(
                im_points_view[start_block:end_block], coa_proj, hae0, tolerance, max_iterations, ref_hae, ref_point)
            start_block = end_block

    if len(orig_shape) == 1:
        coords = numpy.reshape(coords, (-1,))
    elif len(orig_shape) > 1:
        coords = numpy.reshape(coords, orig_shape[:-1] + (3,))
    return coords


#####
# Image-to-DEM

def _do_dem_iteration(
        previous_ecf: numpy.ndarray,
        previous_diff: numpy.ndarray,
        this_ecf: numpy.ndarray,
        this_diff: numpy.ndarray) -> Optional[Tuple[numpy.ndarray, numpy.ndarray]]:
    mask = numpy.isfinite(this_diff) & (this_diff < 0)
    if numpy.any(mask):
        d0 = (previous_diff[mask])
        d1 = numpy.abs(this_diff[mask])
        return mask, (d1[:, numpy.newaxis]*previous_ecf[mask] + d0[:, numpy.newaxis]*this_ecf[mask])/((d0+d1)[:, numpy.newaxis])
    else:
        return None


def _image_to_ground_dem(
        im_points: numpy.ndarray,
        coa_projection: COAProjection,
        dem_interpolator: DEMInterpolator,
        min_dem: float,
        max_dem: float,
        vertical_step_size: Union[float, int],
        ref_hae: float,
        ref_point: numpy.ndarray) -> numpy.ndarray:
    """

    Parameters
    ----------
    im_points : numpy.ndarray
    coa_projection : COAProjection
    dem_interpolator : DEMInterpolator
    min_dem : float
    max_dem : float
    vertical_step_size : float|int
    ref_hae: float
    ref_point : numpy.ndarray

    Returns
    -------
    numpy.ndarray
    """

    # get (image formation specific) projection parameters
    r_tgt_coa, r_dot_tgt_coa, time_coa, arp_coa, varp_coa = coa_projection.projection(im_points)
    ugpn = wgs_84_norm(ref_point)
    tolerance = 1e-3
    max_iterations = 10

    # if max_dem - min_dem is sufficiently small, then pretend it's flat
    if max_dem - min_dem < vertical_step_size:
        return _image_to_ground_hae_perform(
            r_tgt_coa, r_dot_tgt_coa, arp_coa, varp_coa, ref_point, ugpn, max_dem,
            tolerance, max_iterations, ref_hae)
    # set up workspace
    out = numpy.zeros((im_points.shape[0], 3), dtype='float64')
    cont_mask = numpy.ones((im_points.shape[0], ), dtype='bool')
    cont = True
    this_hae = max_dem
    previous_coords = _image_to_ground_hae_perform(
        r_tgt_coa, r_dot_tgt_coa, arp_coa, varp_coa, ref_point, ugpn, max_dem,
        tolerance, max_iterations, ref_hae)
    previous_llh = ecf_to_geodetic(previous_coords)
    previous_diff = previous_llh[:, 2] - dem_interpolator.get_elevation_hae(
        previous_llh[:, 0], previous_llh[:, 1])

    while cont:
        this_hae -= vertical_step_size
        this_coords = _image_to_ground_hae_perform(
            r_tgt_coa[cont_mask], r_dot_tgt_coa[cont_mask], arp_coa[cont_mask], varp_coa[cont_mask],
            ref_point, ugpn, this_hae, tolerance, max_iterations, ref_hae)
        this_llh = ecf_to_geodetic(this_coords)
        this_diff = this_llh[:, 2] - dem_interpolator.get_elevation_hae(this_llh[:, 0], this_llh[:, 1])
        result = _do_dem_iteration(previous_coords, previous_diff, this_coords, this_diff)
        if result is not None:
            this_mask, this_result = result
            temp_mask = numpy.zeros((im_points.shape[0], ), dtype='bool')
            temp_mask[cont_mask] = this_mask
            out[temp_mask, :] = this_result
            cont_mask[temp_mask] = False
            cont = numpy.any(cont_mask)
            if cont:
                previous_coords = this_coords[~this_mask, :]
                previous_diff = this_diff[~this_mask]
        else:
            previous_coords = this_coords
            previous_diff = this_diff
    return out


def _image_to_ground_dem_block(
        im_points: numpy.ndarray,
        coa_projection: COAProjection,
        dem_interpolator: DEMInterpolator,
        horizontal_step: float,
        lat_lon_box: numpy.ndarray,
        block_size: Optional[int],
        lat_pad: float,
        lon_pad: float) -> numpy.ndarray:
    """

    Parameters
    ----------
    im_points : numpy.ndarray
    coa_projection : COAProjection
    dem_interpolator : DEMInterpolator
    horizontal_step : float
    lat_lon_box : numpy.ndarray
    block_size : int|None
    lat_pad : float
    lon_pad : float

    Returns
    -------
    numpy.ndarray
    """

    # determine reference point
    ref_lat = 0.5*(lat_lon_box[0] + lat_lon_box[1])
    ref_lon = 0.5*(lat_lon_box[2] + lat_lon_box[3])
    ref_hae = float(dem_interpolator.get_elevation_hae(ref_lat, ref_lon))
    ref_ecf = geodetic_to_ecf([ref_lat, ref_lon, ref_hae])

    # determine max/min hae in the DEM region
    padded_box = numpy.array([
        max(-90, lat_lon_box[0] - 0.5*lat_pad), min(lat_lon_box[1] + 0.5*lat_pad, 90),
        max(-180, lat_lon_box[2] - 0.5*lon_pad), min(lat_lon_box[3] + 0.5*lon_pad, 180)], dtype='float64')
    min_dem = dem_interpolator.get_min_hae(padded_box) - 10
    max_dem = dem_interpolator.get_max_hae(padded_box) + 10

    # prepare workspace
    num_points = im_points.shape[0]
    if block_size is None or num_points <= block_size:
        coords = _image_to_ground_dem(
            im_points, coa_projection, dem_interpolator, min_dem, max_dem,
            horizontal_step, ref_hae, ref_ecf)
    else:
        coords = numpy.zeros((num_points, 3), dtype='float64')
        # proceed with block processing
        start_block = 0
        while start_block < num_points:
            end_block = min(start_block + block_size, num_points)
            coords[start_block:end_block, :] = _image_to_ground_dem(
                im_points[start_block:end_block, :], coa_projection, dem_interpolator,
                min_dem, max_dem, horizontal_step, ref_hae, ref_ecf)
            start_block = end_block
    return coords


def image_to_ground_dem(
        im_points: Union[numpy.ndarray, list, tuple],
        structure,
        block_size: Optional[int] = 50000,
        dem_interpolator: Union[str, DEMInterpolator] = None,
        dem_type: Union[None, str, List[str]] = None,
        geoid_file: Union[None, str, GeoidHeight] = None,
        pad_value: float = 0.2,
        vertical_step_size: Union[int, float] = 10,
        use_structure_coa: bool = True,
        **coa_args) -> numpy.ndarray:
    """
    Transforms image coordinates to ground plane ECF coordinate via the algorithm(s)
    described in SICD Image Projections document.

    Parameters
    ----------
    im_points : numpy.ndarray|list|tuple
        the image coordinate array
    structure : SICDType|SIDDType
        The SICD or SIDD structure.
    block_size : None|int
        Size of blocks of coordinates to transform at a time. The entire array
        will be transformed as a single block if `None`.
    dem_interpolator : str|DEMInterpolator
        The DEMInterpolator. If this is a string, then a DTEDInterpolator will be
        constructed assuming that this is the DTED root search directory.
    dem_type : None|str|List[str]
        The DEM type or list of DEM types in order of priority. Only used if
        `dem_interpolator` is the search path.
    geoid_file : None|str|GeoidHeight
        The `GeoidHeight` object, an egm file name, or root directory containing
        one of the egm files in the subdirectory "geoid". If `None`, then default
        to the root directory of `dted_list`. Only used if `dem_interpolator` is
        the search path.
    pad_value : float
        The degree value to pad by for the dem interpolator. Only used if
        `dem_interpolator` is the search path.
    vertical_step_size : float|int
        Sampling along HAE altitude at the given resolution in meters. Bounds of
        `[0.1, 100]` will be enforced by replacement.
    use_structure_coa : bool
        If structure.coa_projection is populated, use that one **ignoring the COAProjection parameters.**
    coa_args
        keyword arguments for COAProjection.from_sicd class method.

    Returns
    -------
    numpy.ndarray
        Physical coordinates (in ECF coordinates) with corresponding to the input image
        coordinates, assuming detected features actually correspond to the DEM.
    """

    def append_grid_elements(this_lon_min, this_lon_max, the_list):
        assert this_lon_min <= this_lon_max
        if this_lon_min == this_lon_max:
            if lat_min == lat_max:
                the_list.append((lat_min, lat_max, this_lon_min, this_lon_max))
            else:
                lat_start = lat_min
                while lat_start < lat_max:
                    lat_end = min(lat_start + lat_grid_size, lat_max)
                    the_list.append((lat_start, lat_end, this_lon_min, this_lon_max))
                    lat_start = lat_end
        else:
            if lat_min == lat_max:
                lon_start = this_lon_min
                while lon_start < this_lon_max:
                    lon_end = min(lon_start + lon_grid_size, this_lon_max)
                    the_list.append((lat_min, lat_max, lon_start, lon_end))
                    lon_start = lon_end
            else:
                lat_start = lat_min
                while lat_start < lat_max:
                    lon_start = this_lon_min
                    lat_end = min(lat_start + lat_grid_size, lat_max)
                    while lon_start < this_lon_max:
                        lon_end = min(lon_start + lon_grid_size, this_lon_max)
                        the_list.append((lat_start, lat_end, lon_start, lon_end))
                        lon_start = lon_end
                    lat_start = lat_end

    # coa projection creation
    im_points, orig_shape = _validate_im_points(im_points)
    coa_proj = _get_coa_projection(structure, use_structure_coa, **coa_args)
    vertical_step_size = float(vertical_step_size)
    if vertical_step_size < 0.1:
        vertical_step_size = 0.1
    if vertical_step_size > 100:
        vertical_step_size = 100

    # reference point extraction
    ref_ecf = _get_reference_point(structure)
    ref_llh = ecf_to_geodetic(ref_ecf)
    ref_hae = ref_llh[2]
    # subgrid size definition
    lat_grid_size = 0.03
    lon_grid_size = min(10, lat_grid_size/abs(numpy.sin(numpy.deg2rad(ref_llh[0]))))

    # validate the dem_interpolator
    if dem_interpolator is None:
        raise ValueError('dem_interpolator is None, this is unhandled.')
    if isinstance(dem_interpolator, str):
        dted_list = DTEDList(dem_interpolator)
        dem_interpolator = DTEDInterpolator.from_reference_point(
            ref_llh, dted_list, dem_type=dem_type, geoid_file=geoid_file, pad_value=pad_value)
    if not isinstance(dem_interpolator, DEMInterpolator):
        raise TypeError('dem_interpolator is of unsupported type {}'.format(type(dem_interpolator)))

    # perform a projection to reference point hae for approximate lat/lon values
    im_points_view = numpy.reshape(im_points, (-1, 2))  # possibly or make 2-d flatten
    r_tgt_coa, r_dot_tgt_coa, time_coa, arp_coa, varp_coa = coa_proj.projection(im_points_view)
    ugpn = wgs_84_norm(ref_ecf)
    tolerance = 1e-3
    max_iterations = 10
    llh_rough = ecf_to_geodetic(_image_to_ground_hae_perform(
            r_tgt_coa, r_dot_tgt_coa, arp_coa, varp_coa, ref_ecf, ugpn, ref_hae,
            tolerance, max_iterations, ref_hae))

    # segment into lat/lon grid of small size for more efficient dem lookup
    lat_min = numpy.min(llh_rough[:, 0])
    lat_max = numpy.max(llh_rough[:, 0])
    lon_min = numpy.min(llh_rough[:, 1])
    lon_max = numpy.max(llh_rough[:, 1])
    lat_lon_grids = []
    if (lon_min < -90) and (lon_max > 90):
        # there is a -180/180 crossing
        append_grid_elements(numpy.min(llh_rough[(llh_rough[:, 1] > 0), 1]), 180, lat_lon_grids)
        append_grid_elements(-180, numpy.max(llh_rough[(llh_rough[:, 1] < 0), 1]), lat_lon_grids)
    else:
        append_grid_elements(lon_min, lon_max, lat_lon_grids)

    if len(lat_lon_grids) == 1:
        coords = _image_to_ground_dem_block(
            im_points, coa_proj, dem_interpolator, vertical_step_size,
            lat_lon_grids[0], block_size, lat_grid_size, lon_grid_size)
    else:
        num_points = im_points_view.shape[0]
        coords = numpy.zeros((num_points, 3), dtype='float64')
        for entry in lat_lon_grids:
            mask = ((llh_rough[:, 0] >= entry[0]) & (llh_rough[:, 0] <= entry[1]) &
                    (llh_rough[:, 1] >= entry[2]) & (llh_rough[:, 1] <= entry[3]))
            if numpy.any(mask):
                coords[mask, :] = _image_to_ground_dem_block(
                    im_points_view[mask, :], coa_proj, dem_interpolator, vertical_step_size,
                    entry, block_size, lat_grid_size, lon_grid_size)
    if len(orig_shape) == 1:
        coords = numpy.reshape(coords, (-1,))
    elif len(orig_shape) > 1:
        coords = numpy.reshape(coords, orig_shape[:-1] + (3,))
    return coords


def image_to_slant_sensitivity(sicd_meta, delta_xrow, delta_ycol):
    """
    Compute the sensitivity matrix that relates a change in image location to a change in slant plane.

    The implementation is currently limited to sensitivity around the SCP.

    Parameters
    ----------
    sicd_meta : sarpy.io.complex.sicd_elements.SICD.SICDType
        SICD metadata structure.
    delta_xrow, delta_ycol : array_like
        Increments in xrow/ycol coordinates respectively for which to compute the increments in the slant plane.

    Returns
    -------
    numpy.ndarray
        2x2 sensitivity matrix
    """
    image_coords = numpy.array([
        [0, 0],
        [delta_xrow, 0],
        [0, delta_ycol],
    ])

    image_indices = (image_coords / numpy.array([sicd_meta.Grid.Row.SS, sicd_meta.Grid.Col.SS])[numpy.newaxis, :]
                     + sicd_meta.ImageData.SCPPixel.get_array())

    scp_ecef = sicd_meta.GeoData.SCP.ECF.get_array()
    r, rdot, tcoa, arp, varp = COAProjection.from_sicd(sicd_meta).projection(image_indices)

    vm0 = numpy.linalg.norm(varp[0])
    u_vm = varp[0] / vm0
    u_spx = (arp[0] - scp_ecef) / numpy.linalg.norm(arp[0] - scp_ecef)
    look = {'L': 1, 'R': -1}[sicd_meta.SCPCOA.SideOfTrack]
    spz = look * (numpy.cross(u_spx, u_vm))
    u_spz = spz / numpy.linalg.norm(spz)
    u_spy = numpy.cross(u_spz, u_spx)
    u_vc = numpy.cross(u_spz, u_vm)

    cos_dca0 = -rdot[0] / vm0
    sin_dca0 = numpy.sqrt(1 - cos_dca0**2)

    delta_arp = arp[1:] - arp[0]
    delta_varp = varp[1:] - varp[0]
    delta_r = r[1:] - r[0]
    delta_rdot = rdot[1:] - rdot[0]

    delta_arp_components = delta_arp @ numpy.stack((u_spx, u_spy), axis=-1)
    delta_vm = delta_varp @ u_vm
    delta_vc = delta_varp @ u_vc

    delta_cos_dca = (-1/vm0) * (delta_rdot + delta_vm * (-rdot[0]/vm0))
    delta_dca = (-1/sin_dca0) * delta_cos_dca

    delta_vdir = delta_vc / vm0
    delta_ang = delta_vdir + look * delta_dca

    delta_spx = delta_arp_components[..., 0] - delta_r
    delta_spy = delta_arp_components[..., 1] - r[0] * delta_ang

    return numpy.stack((delta_spx, delta_spy), axis=0) / [delta_xrow, delta_ycol]




from ..tre_elements import TREExtension, TREElement

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


class PT(TREElement):
    def __init__(self, value):
        super(PT, self).__init__()
        self.add_field('LON', 's', 15, value)
        self.add_field('LAT', 's', 15, value)


class BP(TREElement):
    def __init__(self, value):
        super(BP, self).__init__()
        self.add_field('NUM_PTS', 'd', 3, value)
        self.add_loop('PTs', self.NUM_PTS, PT, value)


class MI(TREElement):
    def __init__(self, value):
        super(MI, self).__init__()
        self.add_field('CDV30', 's', 8, value)
        self.add_field('UNIRAT', 's', 3, value)
        self.add_field('RAT', 's', 8, value)
        self.add_field('UNIGMA', 's', 3, value)
        self.add_field('GMA', 's', 8, value)
        self.add_field('LONGMA', 's', 15, value)
        self.add_field('LATGMA', 's', 15, value)
        self.add_field('UNIGCA', 's', 3, value)
        self.add_field('GCA', 's', 8, value)


class LI(TREElement):
    def __init__(self, value):
        super(LI, self).__init__()
        self.add_field('BAD', 's', 10, value)


class PRJ(TREElement):
    def __init__(self, value):
        super(PRJ, self).__init__()
        self.add_field('PRJ', 's', 15, value)


class IN(TREElement):
    def __init__(self, value):
        super(IN, self).__init__()
        self.add_field('INT', 's', 10, value)
        self.add_field('INS_SCA', 's', 9, value)
        self.add_field('NTL', 's', 15, value)
        self.add_field('TTL', 's', 15, value)
        self.add_field('NVL', 's', 15, value)
        self.add_field('TVL', 's', 15, value)
        self.add_field('NTR', 's', 15, value)
        self.add_field('TTR', 's', 15, value)
        self.add_field('NVR', 's', 15, value)
        self.add_field('TVR', 's', 15, value)
        self.add_field('NRL', 's', 15, value)
        self.add_field('TRL', 's', 15, value)
        self.add_field('NSL', 's', 15, value)
        self.add_field('TSL', 's', 15, value)
        self.add_field('NRR', 's', 15, value)
        self.add_field('TRR', 's', 15, value)
        self.add_field('NSR', 's', 15, value)
        self.add_field('TSR', 's', 15, value)


class SOUR(TREElement):
    def __init__(self, value):
        super(SOUR, self).__init__()
        self.add_field('NUM_BP', 'd', 2, value)
        self.add_loop('BPs', self.NUM_BP, BP, value)
        self.add_field('PRT', 's', 10, value)
        self.add_field('URF', 's', 20, value)
        self.add_field('EDN', 's', 7, value)
        self.add_field('NAM', 's', 20, value)
        self.add_field('CDP', 's', 3, value)
        self.add_field('CDV', 's', 8, value)
        self.add_field('CDV27', 's', 8, value)
        self.add_field('SRN', 's', 80, value)
        self.add_field('SCA', 's', 9, value)
        self.add_field('UNISQU', 's', 3, value)
        self.add_field('SQU', 's', 10, value)
        self.add_field('UNIPCI', 's', 3, value)
        self.add_field('PCI', 's', 4, value)
        self.add_field('WPC', 's', 3, value)
        self.add_field('NST', 's', 3, value)
        self.add_field('UNIHKE', 's', 3, value)
        self.add_field('HKE', 's', 6, value)
        self.add_field('LONHKE', 's', 15, value)
        self.add_field('LATHKE', 's', 15, value)
        self.add_field('QSS', 's', 1, value)
        self.add_field('QOD', 's', 1, value)
        self.add_field('CDV10', 's', 8, value)
        self.add_field('QLE', 's', 80, value)
        self.add_field('CPY', 's', 80, value)
        self.add_field('NMI', 'd', 2, value)
        self.add_loop('MIs', self.NMI, MI, value)
        self.add_field('NLI', 'd', 2, value)
        self.add_loop('LIs', self.NLI, LI, value)
        self.add_field('DAG', 's', 80, value)
        self.add_field('DCD', 's', 4, value)
        self.add_field('ELL', 's', 80, value)
        self.add_field('ELC', 's', 3, value)
        self.add_field('DVR', 's', 80, value)
        self.add_field('VDCDVR', 's', 4, value)
        self.add_field('SDA', 's', 80, value)
        self.add_field('VDCSDA', 's', 4, value)
        self.add_field('PRN', 's', 80, value)
        self.add_field('PCO', 's', 2, value)
        self.add_field('NUM_PRJ', 'd', 1, value)
        self.add_loop('PRJs', self.NUM_PRJ, PRJ, value)
        self.add_field('XOR', 's', 15, value)
        self.add_field('YOR', 's', 15, value)
        self.add_field('GRD', 's', 3, value)
        self.add_field('GRN', 's', 80, value)
        self.add_field('ZNA', 's', 4, value)
        self.add_field('NIN', 'd', 2, value)
        self.add_loop('INs', self.NIN, IN, value)


class SOURCBType(TREElement):
    def __init__(self, value):
        super(SOURCBType, self).__init__()
        self.add_field('IS_SCA', 's', 9, value)
        self.add_field('CPATCH', 's', 10, value)
        self.add_field('NUM_SOUR', 'd', 2, value)
        self.add_loop('SOURs', self.NUM_SOUR, SOUR, value)


class SOURCB(TREExtension):
    _tag_value = 'SOURCB'
    _data_type = SOURCBType




from ..tre_elements import TREExtension, TREElement

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"

# TODO: I'm not entirely confident that these bit operation work as expected


class PARAMETER(TREElement):
    def __init__(self, value, EXISTENCE_MASK):
        super(PARAMETER, self).__init__()
        if EXISTENCE_MASK & 0x10000000:
            self.add_field('BANDID', 's', 50, value)
        if EXISTENCE_MASK & 0x08000000:
            self.add_field('BAD_BAND', 'd', 1, value)
        if EXISTENCE_MASK & 0x04000000:
            self.add_field('NIIRS', 's', 3, value)
        if EXISTENCE_MASK & 0x02000000:
            self.add_field('FOCAL_LEN', 'd', 5, value)
        if EXISTENCE_MASK & 0x01000000:
            self.add_field('CWAVE', 's', 7, value)
        if EXISTENCE_MASK & 0x00800000:
            self.add_field('FWHM', 's', 7, value)
        if EXISTENCE_MASK & 0x00400000:
            self.add_field('FWHM_UNC', 's', 7, value)
        if EXISTENCE_MASK & 0x00200000:
            self.add_field('NOM_WAVE', 's', 7, value)
        if EXISTENCE_MASK & 0x00100000:
            self.add_field('NOM_WAVE_UNC', 's', 7, value)
        if EXISTENCE_MASK & 0x00080000:
            self.add_field('LBOUND', 's', 7, value)
            self.add_field('UBOUND', 's', 7, value)
        if EXISTENCE_MASK & 0x00040000:
            self.add_field('SCALE_FACTOR', 'ieee754_binary32', 4, value)
            self.add_field('ADDITIVE_FACTOR', 'ieee754_binary32', 4, value)
        if EXISTENCE_MASK & 0x00020000:
            self.add_field('START_TIME', 's', 16, value)
        if EXISTENCE_MASK & 0x00010000:
            self.add_field('INT_TIME', 's', 6, value)
        if EXISTENCE_MASK & 0x00008000:
            self.add_field('CALDRK', 's', 6, value)
            self.add_field('CALIBRATION_SENSITIVITY', 's', 5, value)
        if EXISTENCE_MASK & 0x00004000:
            self.add_field('ROW_GSD', 's', 7, value)
            if EXISTENCE_MASK & 0x00002000:
                self.add_field('ROW_GSD_UNC', 's', 7, value)
            self.add_field('ROW_GSD_UNIT', 's', 1, value)
            self.add_field('COL_GSD', 's', 7, value)
            if EXISTENCE_MASK & 0x00002000:
                self.add_field('COL_GSD_UNC', 's', 7, value)
            self.add_field('COL_GSD_UNIT', 's', 1, value)
        if EXISTENCE_MASK & 0x00001000:
            self.add_field('BKNOISE', 's', 5, value)
            self.add_field('SCNNOISE', 's', 5, value)
        if EXISTENCE_MASK & 0x00000800:
            self.add_field('SPT_RESP_FUNCTION_ROW', 's', 7, value)
            if EXISTENCE_MASK & 0x00000400:
                self.add_field('SPT_RESP_UNC_ROW', 's', 7, value)
            self.add_field('SPT_RESP_UNIT_ROW', 's', 1, value)
            self.add_field('SPT_RESP_FUNCTION_COL', 's', 7, value)
            if EXISTENCE_MASK & 0x00000400:
                self.add_field('SPT_RESP_UNC_COL', 's', 7, value)
            self.add_field('SPT_RESP_UNIT_COL', 's', 1, value)
        if EXISTENCE_MASK & 0x00000200:
            self.add_field('DATA_FLD_3', 'b', 16, value)
        if EXISTENCE_MASK & 0x00000100:
            self.add_field('DATA_FLD_4', 'b', 24, value)
        if EXISTENCE_MASK & 0x00000080:
            self.add_field('DATA_FLD_5', 'b', 32, value)
        if EXISTENCE_MASK & 0x00000040:
            self.add_field('DATA_FLD_6', 'b', 48, value)


class BAND(TREElement):
    def __init__(self, value):
        super(BAND, self).__init__()
        if self.BAPF == 'I':
            self.add_field('APN', 'd', 10, value)
        if self.BAPF == 'R':
            self.add_field('APR', 'ieee754_binary32', 4, value)
        if self.BAPF == 'A':
            self.add_field('APA', 's', 20, value)


class AUX_B(TREElement):
    def __init__(self, value):
        super(AUX_B, self).__init__()
        self.add_field('BAPF', 's', 1, value)
        self.add_field('UBAP', 's', 7, value)
        self.add_loop('BANDs', self.COUNT, BAND, value)


class AUX_C(TREElement):
    def __init__(self, value):
        super(AUX_C, self).__init__()
        self.add_field('CAPF', 's', 1, value)
        self.add_field('UCAP', 's', 7, value)
        if self.CAPF == 'I':
            self.add_field('APN', 'd', 10, value)
        if self.CAPF == 'R':
            self.add_field('APR', 'ieee754_binary32', 4, value)
        if self.CAPF == 'A':
            self.add_field('APA', 's', 20, value)


class BANDSBType(TREElement):
    def __init__(self, value):
        super(BANDSBType, self).__init__()
        self.add_field('COUNT', 'd', 5, value)
        self.add_field('RADIOMETRIC_QUANTITY', 's', 24, value)
        self.add_field('RADIOMETRIC_QUANTITY_UNIT', 's', 1, value)
        self.add_field('SCALE_FACTOR', 'ieee754_binary32', 4, value)
        self.add_field('ADDITIVE_FACTOR', 'ieee754_binary32', 4, value)
        self.add_field('ROW_GSD', 's', 7, value)
        self.add_field('ROW_GSD_UNIT', 's', 1, value)
        self.add_field('COL_GSD', 's', 7, value)
        self.add_field('COL_GSD_UNIT', 's', 1, value)
        self.add_field('SPT_RESP_ROW', 's', 7, value)
        self.add_field('SPT_RESP_UNIT_ROW', 's', 1, value)
        self.add_field('SPT_RESP_COL', 's', 7, value)
        self.add_field('SPT_RESP_UNIT_COL', 's', 1, value)
        self.add_field('DATA_FLD_1', 'b', 48, value)
        self.add_field('EXISTENCE_MASK', 'b', 4, value)
        existence_mask = int.from_bytes(self.EXISTENCE_MASK, byteorder='big')
        if existence_mask & 0x80000000:
            self.add_field('RADIOMETRIC_ADJUSTMENT_SURFACE', 's', 24, value)
            self.add_field('ATMOSPHERIC_ADJUSTMENT_ALTITUDE', 'ieee754_binary32', 4, value)
        if existence_mask & 0x40000000:
            self.add_field('DIAMETER', 's', 7, value)
        if existence_mask & 0x20000000:
            self.add_field('DATA_FLD_2', 'b', 32, value)
        if existence_mask & 0x01F80000:
            self.add_field('WAVE_LENGTH_UNIT', 's', 1, value)
        self.add_loop('PARAMETERs', self.COUNT, PARAMETER, value, existence_mask)
        if existence_mask & 0x00000001:
            self.add_field('NUM_AUX_B', 'd', 2, value)
            self.add_field('NUM_AUX_C', 'd', 2, value)
            self.add_loop('AUX_Bs', self.NUM_AUX_B, AUX_B, value)
            self.add_loop('AUX_Cs', self.NUM_AUX_C, AUX_C, value)


class BANDSB(TREExtension):
    _tag_value = 'BANDSB'
    _data_type = BANDSBType



"""
The reference geometry parameters definition.
"""

__classification__ = "UNCLASSIFIED"
__author__ = ("Thomas McCullough", "Michael Stewart, Valkyrie")


import numpy

from sarpy.io.xml.base import Serializable, parse_serializable
from sarpy.io.xml.descriptors import FloatDescriptor, StringEnumDescriptor, \
    SerializableDescriptor
from sarpy.io.complex.sicd_elements.blocks import XYZType, LatLonHAEType
from sarpy.geometry.geocoords import geodetic_to_ecf, ecf_to_geodetic

from .base import DEFAULT_STRICT, FLOAT_FORMAT


class CRPType(Serializable):
    """
    The CRP position for the reference vector of the reference channel.
    """

    _fields = ('ECF', 'LLH')
    _required = _fields
    _ECF = None
    _LLH = None

    def __init__(self, ECF=None, LLH=None, **kwargs):
        """
        To avoid the potential of inconsistent state, ECF and LLH are not simultaneously
        used. If ECF is provided, it is used to populate LLH. Otherwise, if LLH is provided,
        then it is used the populate ECF.

        Parameters
        ----------
        ECF : XYZType|numpy.ndarray|list|tuple
        LLH : LatLonHAEType|numpy.ndarray|list|tuple
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        if ECF is not None:
            self.ECF = ECF
        elif LLH is not None:
            self.LLH = LLH
        super(CRPType, self).__init__(**kwargs)

    @property
    def ECF(self):  # type: () -> XYZType
        """
        XYZType: The CRP Position ECF coordinates.
        """

        return self._ECF

    @ECF.setter
    def ECF(self, value):
        if value is not None:
            self._ECF = parse_serializable(value, 'ECF', self, XYZType)
            self._LLH = LatLonHAEType.from_array(ecf_to_geodetic(self._ECF.get_array()))

    @property
    def LLH(self):  # type: () -> LatLonHAEType
        """
        LatLonHAEType: The CRP Position in WGS-84 coordinates.
        """

        return self._LLH

    @LLH.setter
    def LLH(self, value):
        if value is not None:
            self._LLH = parse_serializable(value, 'LLH', self, LatLonHAEType)
            self._ECF = XYZType.from_array(geodetic_to_ecf(self._LLH.get_array(order='LAT')))


class RcvParametersType(Serializable):
    """
    The receive parameters geometry implementation.
    """

    _fields = (
        'RcvTime', 'RcvPos', 'RcvVel', 'SideOfTrack', 'SlantRange', 'GroundRange',
        'DopplerConeAngle', 'GrazeAngle', 'IncidenceAngle', 'AzimuthAngle')
    _required = _fields
    _numeric_format = {
        'SlantRange': FLOAT_FORMAT, 'GroundRange': FLOAT_FORMAT, 'DopplerConeAngle': FLOAT_FORMAT,
        'GrazeAngle': FLOAT_FORMAT, 'IncidenceAngle': FLOAT_FORMAT, 'AzimuthAngle': FLOAT_FORMAT}
    # descriptors
    RcvTime = FloatDescriptor(
        'RcvTime', _required, strict=DEFAULT_STRICT,
        docstring='Receive time for the first sample for the reference vector.')  # type: float
    RcvPos = SerializableDescriptor(
        'RcvPos', XYZType, _required, strict=DEFAULT_STRICT,
        docstring='APC position in ECF coordinates.')  # type: XYZType
    RcvVel = SerializableDescriptor(
        'RcvVel', XYZType, _required, strict=DEFAULT_STRICT,
        docstring='APC velocity in ECF coordinates.')  # type: XYZType
    SideOfTrack = StringEnumDescriptor(
        'SideOfTrack', ('L', 'R'), _required, strict=DEFAULT_STRICT,
        docstring='Side of Track parameter for the collection.')  # type: str
    SlantRange = FloatDescriptor(
        'SlantRange', _required, strict=DEFAULT_STRICT, bounds=(0, None),
        docstring='Slant range from the APC to the CRP.')  # type: float
    GroundRange = FloatDescriptor(
        'GroundRange', _required, strict=DEFAULT_STRICT, bounds=(0, None),
        docstring='Ground range from the APC nadir to the CRP.')  # type: float
    DopplerConeAngle = FloatDescriptor(
        'DopplerConeAngle', _required, strict=DEFAULT_STRICT, bounds=(0, 180),
        docstring='Doppler Cone Angle between APC velocity and deg CRP Line of '
                  'Sight (LOS).')  # type: float
    GrazeAngle = FloatDescriptor(
        'GrazeAngle', _required, strict=DEFAULT_STRICT, bounds=(0, 90),
        docstring='Grazing angle for the APC to CRP LOS and the Earth Tangent '
                  'Plane (ETP) at the CRP.')  # type: float
    IncidenceAngle = FloatDescriptor(
        'IncidenceAngle', _required, strict=DEFAULT_STRICT, bounds=(0, 90),
        docstring='Incidence angle for the APC to CRP LOS and the Earth Tangent '
                  'Plane (ETP) at the CRP.')  # type: float
    AzimuthAngle = FloatDescriptor(
        'AzimuthAngle', _required, strict=DEFAULT_STRICT, bounds=(0, 360),
        docstring='Angle from north to the line from the CRP to the APC ETP '
                  'Nadir (i.e. North to +GPX). Measured clockwise from North '
                  'toward East.')  # type: float

    def __init__(self, RcvTime=None, RcvPos=None, RcvVel=None,
                 SideOfTrack=None, SlantRange=None, GroundRange=None,
                 DopplerConeAngle=None, GrazeAngle=None, IncidenceAngle=None,
                 AzimuthAngle=None, **kwargs):
        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.RcvTime = RcvTime
        self.RcvPos = RcvPos
        self.RcvVel = RcvVel
        self.SideOfTrack = SideOfTrack
        self.SlantRange = SlantRange
        self.GroundRange = GroundRange
        self.DopplerConeAngle = DopplerConeAngle
        self.GrazeAngle = GrazeAngle
        self.IncidenceAngle = IncidenceAngle
        self.AzimuthAngle = AzimuthAngle
        super(RcvParametersType, self).__init__(**kwargs)

    @property
    def look(self):
        """
        int: An integer version of `SideOfTrack`:

            * None if `SideOfTrack` is not defined

            * -1 if SideOfTrack == 'R'

            * 1 if SideOftrack == 'L'
        """

        if self.SideOfTrack is None:
            return None
        return -1 if self.SideOfTrack == 'R' else 1


class ReferenceGeometryType(Serializable):
    """
    Parameters that describe the collection geometry for the reference vector
    of the reference channel.
    """

    _fields = ('CRP', 'RcvParameters')
    _required = _fields
    # descriptors
    CRP = SerializableDescriptor(
        'CRP', CRPType, _required, strict=DEFAULT_STRICT,
        docstring='The Collection Reference Point (CRP) used for computing the'
                  ' geometry parameters.')  # type: CRPType
    RcvParameters = SerializableDescriptor(
        'RcvParameters', RcvParametersType, _required, strict=DEFAULT_STRICT,
        docstring='Parameters computed for the receive APC position.')  # type: RcvParametersType

    def __init__(self, CRP=None, RcvParameters=None, **kwargs):
        """

        Parameters
        ----------
        CRP : CRPType
        RcvParameters : RcvParametersType
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.CRP = CRP
        self.RcvParameters = RcvParameters
        super(ReferenceGeometryType, self).__init__(**kwargs)



"""
Base common features for complex readers
"""


__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"

from typing import Union, Tuple, Sequence, Callable
import numpy

from sarpy.io.complex.sicd_elements.SICD import SICDType
from sarpy.io.complex.sicd_elements.utils import is_general_match
from sarpy.io.general.base import BaseReader, FlatReader
from sarpy.io.general.data_segment import DataSegment, SubsetSegment
from sarpy.io.general.format_function import FormatFunction


class SICDTypeReader(BaseReader):
    """
    A class for ensuring common SICD reading functionality.

    **Changed in version 1.3.0** for reading changes.
    """

    def __init__(self,
                 data_segment: Union[None, DataSegment, Sequence[DataSegment]],
                 sicd_meta: Union[None, SICDType, Sequence[SICDType]],
                 close_segments: bool = True,
                 delete_files: Union[None, str, Sequence[str]] = None):
        """

        Parameters
        ----------
        data_segment : None|DataSegment|Sequence[DataSegment]
        sicd_meta : None|SICDType|Sequence[SICDType]
            The SICD metadata object(s).
        close_segments : bool
            Call segment.close() for each data segment on reader.close()?
        delete_files : None|Sequence[str]
            Any temp files which should be cleaned up on reader.close()?
            This will occur after closing segments.
        """

        if sicd_meta is None:
            self._sicd_meta = None
        elif isinstance(sicd_meta, SICDType):
            self._sicd_meta = sicd_meta
        else:
            temp_list = []
            for el in sicd_meta:
                if not isinstance(el, SICDType):
                    raise TypeError(
                        'Got a collection for sicd_meta, and all elements are required '
                        'to be instances of SICDType.')
                temp_list.append(el)
            self._sicd_meta = tuple(temp_list)

        BaseReader.__init__(
            self, data_segment, reader_type='SICD', close_segments=close_segments, delete_files=delete_files)

    def _check_sizes(self) -> None:
        data_sizes = self.get_data_size_as_tuple()
        sicds = self.get_sicds_as_tuple()
        agree = True
        msg = ''
        for i, (data_size, sicd) in enumerate(zip(data_sizes, sicds)):
            expected_size = (sicd.ImageData.NumRows, sicd.ImageData.NumCols)
            if data_size != expected_size:
                agree = False
                msg += 'data segment at index {} has data size {}\n\t' \
                       'and expected size (from the sicd) {}\n'.format(i, data_size, expected_size)
        if not agree:
            raise ValueError(msg)

    @property
    def sicd_meta(self) -> Union[None, SICDType, Tuple[SICDType, ...]]:
        """
        None|SICDType|Tuple[SICDType, ...]: the sicd meta_data or meta_data collection.
        """

        return self._sicd_meta

    def get_sicds_as_tuple(self) -> Union[None, Tuple[SICDType, ...]]:
        """
        Get the sicd or sicd collection as a tuple - for simplicity and consistency of use.

        Returns
        -------
        None|Tuple[SICDType, ...]
        """

        if self.sicd_meta is None:
            return None
        elif isinstance(self.sicd_meta, tuple):
            return self.sicd_meta
        else:
            # noinspection PyRedundantParentheses
            return (self.sicd_meta, )

    def get_sicd_partitions(self, match_function: Callable = is_general_match) -> Tuple[Tuple[int, ...], ...]:
        """
        Partition the sicd collection into sub-collections according to `match_function`,
        which is assumed to establish an equivalence relation.

        Parameters
        ----------
        match_function : callable
            This match function must have call signature `(SICDType, SICDType) -> bool`, and
            defaults to :func:`sarpy.io.complex.sicd_elements.utils.is_general_match`.
            This function is assumed reflexive, symmetric, and transitive.

        Returns
        -------
        Tuple[Tuple[int, ...], ...]
        """

        sicds = self.get_sicds_as_tuple()
        # set up or state workspace
        count = len(sicds)
        matched = numpy.zeros((count,), dtype='bool')
        matches = []

        # assemble or match collections
        for i in range(count):
            if matched[i]:
                # it's already matched somewhere
                continue

            matched[i] = True  # shouldn't access backwards, but just to be thorough
            this_match = [i, ]
            for j in range(i + 1, count):
                if not matched[j] and match_function(sicds[i], sicds[j]):
                    matched[j] = True
                    this_match.append(j)
            matches.append(tuple(this_match))
        return tuple(matches)

    def get_sicd_bands(self) -> Tuple[str, ...]:
        """
        Gets the list of bands for each sicd.

        Returns
        -------
        Tuple[str, ...]
        """

        return tuple(sicd.get_transmit_band_name() for sicd in self.get_sicds_as_tuple())

    def get_sicd_polarizations(self) -> Tuple[str, ...]:
        """
        Gets the list of polarizations for each sicd.

        Returns
        -------
        Tuple[str]
        """

        return tuple(sicd.get_processed_polarization() for sicd in self.get_sicds_as_tuple())


class FlatSICDReader(FlatReader, SICDTypeReader):
    """
    Create a sicd type reader directly from an array.

    **Changed in version 1.3.0** for reading changes.
    """

    def __init__(self,
                 sicd_meta,
                 underlying_array,
                 formatted_dtype: Union[None, str, numpy.dtype] = None,
                 formatted_shape: Union[None, Tuple[int, ...]] = None,
                 reverse_axes: Union[None, int, Sequence[int]] = None,
                 transpose_axes: Union[None, Tuple[int, ...]] = None,
                 format_function: Union[None, FormatFunction] = None,
                 close_segments: bool = True):
        """

        Parameters
        ----------
        sicd_meta : None|SICDType
            `None`, or the SICD metadata object
        underlying_array : numpy.ndarray
        formatted_dtype : None|str|numpy.dtype
        formatted_shape : None|Tuple[int, ...]
        reverse_axes : None|Sequence[int]
        transpose_axes : None|Tuple[int, ...]
        format_function : None|FormatFunction
        close_segments : bool
        """

        FlatReader.__init__(
            self, underlying_array,
            formatted_dtype=formatted_dtype, formatted_shape=formatted_shape,
            reverse_axes=reverse_axes, transpose_axes=transpose_axes,
            format_function=format_function, close_segments=close_segments)
        SICDTypeReader.__init__(self, None, sicd_meta)
        self._check_sizes()

    def write_to_file(self, output_file, check_older_version=False, check_existence=False):
        """
        Write a file for the given in-memory reader.

        Parameters
        ----------
        output_file : str
        check_older_version : bool
            Try to use a less recent version of SICD (1.1), for possible application compliance issues?
        check_existence : bool
            Should we check if the given file already exists, and raise an exception if so?
        """

        if not isinstance(output_file, str):
            raise TypeError(
                'output_file is expected to a be a string, got type {}'.format(type(output_file)))

        from sarpy.io.complex.sicd import SICDWriter
        with SICDWriter(
                output_file, self.sicd_meta,
                check_older_version=check_older_version, check_existence=check_existence) as writer:
            writer.write_chip(self[:, :], start_indices=(0, 0))


class SubsetSICDReader(SICDTypeReader):
    """
    Create a reader based on a specific subset of a given SICDTypeReader.

    **Changed in version 1.3.0** for reading changes.
    """

    def __init__(self, reader, row_bounds, column_bounds, index=0, close_parent=False):
        """

        Parameters
        ----------
        reader : SICDTypeReader
            The base reader.
        row_bounds : None|Tuple[int, int]
            Of the form `(min row, max row)`.
        column_bounds : None|Tuple[int, int]
            Of the form `(min column, max column)`.
        index : int
            The image index.
        close_parent : bool
        """

        sicd, row_bounds, column_bounds = reader.get_sicds_as_tuple()[index].create_subset_structure(
            row_bounds, column_bounds)

        parent_segment = reader.get_data_segment_as_tuple()[index]
        subset_definition = (slice(*row_bounds), slice(*column_bounds))
        data_segment = SubsetSegment(
            parent_segment, subset_definition, coordinate_basis='formatted', close_parent=close_parent)
        SICDTypeReader.__init__(self, data_segment, sicd)

    @property
    def file_name(self) -> None:
        return None
